import json

from moto.core.exceptions import JsonRESTError


class MQError(JsonRESTError):
    pass


class UnknownBroker(MQError):
    def __init__(self, broker_id: str):
        super().__init__("NotFoundException", "Can't find requested broker")
        body = {
            "errorAttribute": "broker-id",
            "message": f"Can't find requested broker [{broker_id}]. Make sure your broker exists.",
        }
        self.description = json.dumps(body)


class UnknownConfiguration(MQError):
    def __init__(self, config_id: str):
        super().__init__("NotFoundException", "Can't find requested configuration")
        body = {
            "errorAttribute": "configuration_id",
            "message": f"Can't find requested configuration [{config_id}]. Make sure your configuration exists.",
        }
        self.description = json.dumps(body)


class UnknownUser(MQError):
    def __init__(self, username: str):
        super().__init__("NotFoundException", "Can't find requested user")
        body = {
            "errorAttribute": "username",
            "message": f"Can't find requested user [{username}]. Make sure your user exists.",
        }
        self.description = json.dumps(body)


class UnknownEngineType(MQError):
    def __init__(self, engine_type: str):
        super().__init__("BadRequestException", "")
        body = {
            "errorAttribute": "engineType",
            "message": f"Broker engine type [{engine_type}] is invalid. Valid values are: [ACTIVEMQ]",
        }
        self.description = json.dumps(body)



import enum
from typing import Final

from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.eval.environment import Environment


class MaxAttemptsOutcome(enum.Enum):
    EXHAUSTED = False
    SUCCESS = True


class MaxAttemptsDecl(EvalComponent):
    """
    "MaxAttempts": value MUST be a non-negative integer, representing the maximum number
    of retry attempts (default: 3)
    """

    DEFAULT_ATTEMPTS: Final[int] = 3
    MAX_VALUE: Final[int] = 99999999

    attempts: Final[int]

    def __init__(self, attempts: int = DEFAULT_ATTEMPTS):
        if not (1 <= attempts <= MaxAttemptsDecl.MAX_VALUE):
            raise ValueError(
                f"MaxAttempts value MUST be a positive integer between "
                f"1 and {MaxAttemptsDecl.MAX_VALUE}, got '{attempts}'."
            )
        self.attempts = attempts

    def _attempt_number_key(self) -> str:
        return f"MaxAttemptsDecl-{self.heap_key}-attempt_number"

    def _access_attempt_number(self, env: Environment) -> int:
        return env.heap.get(self._attempt_number_key(), -1)

    def _store_attempt_number(self, env: Environment, attempt_number: float) -> None:
        env.heap[self._attempt_number_key()] = attempt_number

    def _eval_body(self, env: Environment) -> None:
        if self.attempts == 0:
            env.stack.append(MaxAttemptsOutcome.SUCCESS)
        else:
            attempt_number: int = self._access_attempt_number(env=env)
            attempt_number += 1
            env.stack.append(MaxAttemptsOutcome(attempt_number < self.attempts))
            self._store_attempt_number(env=env, attempt_number=attempt_number)



from moto.utilities.utils import get_partition


def make_arn(
    region: str, account_id: str, resource_type: str, resource_path: str
) -> str:
    arn_template = "arn:{partition}:elasticbeanstalk:{region}:{account_id}:{resource_type}/{resource_path}"
    arn = arn_template.format(
        partition=get_partition(region),
        region=region,
        account_id=account_id,
        resource_type=resource_type,
        resource_path=resource_path,
    )
    return arn



import pytest
from freezegun import freeze_time

from moto.swf.models import HistoryEvent


@freeze_time("2015-01-01 12:00:00")
def test_history_event_creation():
    he = HistoryEvent(123, "DecisionTaskStarted", scheduled_event_id=2)
    assert he.event_id == 123
    assert he.event_type == "DecisionTaskStarted"
    assert he.event_timestamp == 1420113600.0


@freeze_time("2015-01-01 12:00:00")
def test_history_event_to_dict_representation():
    he = HistoryEvent(123, "DecisionTaskStarted", scheduled_event_id=2)
    assert he.to_dict() == {
        "eventId": 123,
        "eventType": "DecisionTaskStarted",
        "eventTimestamp": 1420113600.0,
        "decisionTaskStartedEventAttributes": {"scheduledEventId": 2},
    }


def test_history_event_breaks_on_initialization_if_not_implemented():
    with pytest.raises(NotImplementedError):
        HistoryEvent(123, "UnknownHistoryEvent")



import json
import re
from typing import Any, Dict, List

from moto.core.responses import BaseResponse
from moto.utilities.utils import get_partition

from .exceptions import (
    InvalidInputException,
    ResourceAlreadyExistsException,
    ResourceNotFoundException,
)
from .models import CodeBuildBackend, codebuild_backends


def _validate_required_params_source(source: Dict[str, Any]) -> None:
    if source["type"] not in [
        "BITBUCKET",
        "CODECOMMIT",
        "CODEPIPELINE",
        "GITHUB",
        "GITHUB_ENTERPRISE",
        "NO_SOURCE",
        "S3",
    ]:
        raise InvalidInputException("Invalid type provided: Project source type")

    if "location" not in source:
        raise InvalidInputException("Project source location is required")

    if source["location"] == "":
        raise InvalidInputException("Project source location is required")


def _validate_required_params_service_role(
    account_id: str, region_name: str, service_role: str
) -> None:
    if not service_role.startswith(
        f"arn:{get_partition(region_name)}:iam::{account_id}:role/"
    ):
        raise InvalidInputException(
            "Invalid service role: Service role account ID does not match caller's account"
        )


def _validate_required_params_artifacts(artifacts: Dict[str, Any]) -> None:
    if artifacts["type"] not in ["CODEPIPELINE", "S3", "NO_ARTIFACTS"]:
        raise InvalidInputException("Invalid type provided: Artifact type")

    if artifacts["type"] == "NO_ARTIFACTS":
        if "location" in artifacts:
            raise InvalidInputException(
                "Invalid artifacts: artifact type NO_ARTIFACTS should have null location"
            )
    elif "location" not in artifacts or artifacts["location"] == "":
        raise InvalidInputException("Project source location is required")


def _validate_required_params_environment(environment: Dict[str, Any]) -> None:
    if environment["type"] not in [
        "WINDOWS_CONTAINER",
        "LINUX_CONTAINER",
        "LINUX_GPU_CONTAINER",
        "ARM_CONTAINER",
    ]:
        raise InvalidInputException(f"Invalid type provided: {environment['type']}")

    if environment["computeType"] not in [
        "BUILD_GENERAL1_SMALL",
        "BUILD_GENERAL1_MEDIUM",
        "BUILD_GENERAL1_LARGE",
        "BUILD_GENERAL1_2XLARGE",
    ]:
        raise InvalidInputException(
            f"Invalid compute type provided: {environment['computeType']}"
        )


def _validate_required_params_project_name(name: str) -> None:
    if len(name) >= 150:
        raise InvalidInputException(
            "Only alphanumeric characters, dash, and underscore are supported"
        )

    if not re.match(r"^[A-Za-z]{1}.*[^!£$%^&*()+=|?`¬{}@~#:;<>\\/\[\]]$", name):
        raise InvalidInputException(
            "Only alphanumeric characters, dash, and underscore are supported"
        )


def _validate_required_params_id(build_id: str, build_ids: List[str]) -> None:
    if ":" not in build_id:
        raise InvalidInputException("Invalid build ID provided")

    if build_id not in build_ids:
        raise ResourceNotFoundException(f"Build {build_id} does not exist")


class CodeBuildResponse(BaseResponse):
    @property
    def codebuild_backend(self) -> CodeBuildBackend:
        return codebuild_backends[self.current_account][self.region]

    def list_builds_for_project(self) -> str:
        _validate_required_params_project_name(self._get_param("projectName"))

        if (
            self._get_param("projectName")
            not in self.codebuild_backend.codebuild_projects.keys()
        ):
            name = self._get_param("projectName")
            raise ResourceNotFoundException(
                f"The provided project arn:{get_partition(self.region)}:codebuild:{self.region}:{self.current_account}:project/{name} does not exist"
            )

        ids = self.codebuild_backend.list_builds_for_project(
            self._get_param("projectName")
        )

        return json.dumps({"ids": ids})

    def create_project(self) -> str:
        _validate_required_params_source(self._get_param("source"))
        service_role = self._get_param("serviceRole")
        _validate_required_params_service_role(
            self.current_account, self.region, service_role
        )
        _validate_required_params_artifacts(self._get_param("artifacts"))
        _validate_required_params_environment(self._get_param("environment"))
        _validate_required_params_project_name(self._get_param("name"))

        if self._get_param("name") in self.codebuild_backend.codebuild_projects.keys():
            name = self._get_param("name")
            raise ResourceAlreadyExistsException(
                f"Project already exists: arn:{get_partition(self.region)}:codebuild:{self.region}:{self.current_account}:project/{name}"
            )

        project_metadata = self.codebuild_backend.create_project(
            self._get_param("name"),
            self._get_param("source"),
            self._get_param("artifacts"),
            self._get_param("environment"),
            service_role=service_role,
        )

        return json.dumps({"project": project_metadata})

    def list_projects(self) -> str:
        project_metadata = self.codebuild_backend.list_projects()
        return json.dumps({"projects": project_metadata})

    def start_build(self) -> str:
        _validate_required_params_project_name(self._get_param("projectName"))

        if (
            self._get_param("projectName")
            not in self.codebuild_backend.codebuild_projects.keys()
        ):
            name = self._get_param("projectName")
            raise ResourceNotFoundException(
                f"Project cannot be found: arn:{get_partition(self.region)}:codebuild:{self.region}:{self.current_account}:project/{name}"
            )

        metadata = self.codebuild_backend.start_build(
            self._get_param("projectName"),
            self._get_param("sourceVersion"),
            self._get_param("artifactsOverride"),
        )
        return json.dumps({"build": metadata})

    def batch_get_builds(self) -> str:
        for build_id in self._get_param("ids"):
            if ":" not in build_id:
                raise InvalidInputException("Invalid build ID provided")

        metadata = self.codebuild_backend.batch_get_builds(self._get_param("ids"))
        return json.dumps({"builds": metadata})

    def list_builds(self) -> str:
        ids = self.codebuild_backend.list_builds()
        return json.dumps({"ids": ids})

    def delete_project(self) -> str:
        _validate_required_params_project_name(self._get_param("name"))

        self.codebuild_backend.delete_project(self._get_param("name"))
        return "{}"

    def stop_build(self) -> str:
        _validate_required_params_id(
            self._get_param("id"), self.codebuild_backend.list_builds()
        )

        metadata = self.codebuild_backend.stop_build(self._get_param("id"))
        return json.dumps({"build": metadata})



from typing import Any

from moto.core.exceptions import RESTError


class SesError(RESTError):
    code = 400

    def __init__(self, error_type: str, message: str, *args: Any, **kwargs: Any):
        kwargs.setdefault("template", "wrapped_single_error")
        super().__init__(error_type, message, *args, **kwargs)


class MessageRejectedError(SesError):
    def __init__(self, message: str):
        super().__init__("MessageRejected", message)


class ConfigurationSetDoesNotExist(SesError):
    def __init__(self, message: str):
        super().__init__("ConfigurationSetDoesNotExist", message)


class ConfigurationSetAlreadyExists(SesError):
    def __init__(self, message: str):
        super().__init__("ConfigurationSetAlreadyExists", message)


class EventDestinationAlreadyExists(SesError):
    def __init__(self, message: str):
        super().__init__("EventDestinationAlreadyExists", message)


class TemplateNameAlreadyExists(SesError):
    def __init__(self, message: str):
        super().__init__("TemplateNameAlreadyExists", message)


class ValidationError(SesError):
    def __init__(self, message: str):
        super().__init__("ValidationError", message)


class InvalidParameterValue(SesError):
    def __init__(self, message: str):
        super().__init__("InvalidParameterValue", message)


class InvalidRenderingParameterException(SesError):
    def __init__(self, message: str):
        super().__init__("InvalidRenderingParameterException", message)


class TemplateDoesNotExist(SesError):
    def __init__(self, message: str):
        super().__init__("TemplateDoesNotExist", message)


class RuleSetNameAlreadyExists(SesError):
    def __init__(self, message: str):
        super().__init__("RuleSetNameAlreadyExists", message)


class RuleAlreadyExists(SesError):
    def __init__(self, message: str):
        super().__init__("RuleAlreadyExists", message)


class RuleSetDoesNotExist(SesError):
    def __init__(self, message: str):
        super().__init__("RuleSetDoesNotExist", message)


class RuleDoesNotExist(SesError):
    def __init__(self, message: str):
        super().__init__("RuleDoesNotExist", message)


class MissingRenderingAttributeException(SesError):
    def __init__(self, var: str):
        super().__init__(
            "MissingRenderingAttributeException",
            f"Attribute '{var}' is not present in the rendering data.",
        )



from time import sleep
from uuid import uuid4

import boto3
import pytest

from . import elbv2_aws_verified


@elbv2_aws_verified()
@pytest.mark.aws_verified
def test_register_targets(target_is_aws=False):
    conn = boto3.client("elbv2", region_name="us-east-1")
    ec2 = boto3.resource("ec2", region_name="us-east-1")
    ec2_client = boto3.client("ec2", region_name="us-east-1")

    lb_name = f"lb-{str(uuid4())[0:6]}"
    sg_name = f"sg{str(uuid4())[0:6]}"
    target_name = f"target-{str(uuid4())[0:6]}"

    ssm = boto3.client("ssm", "us-east-1")
    resp = ssm.get_parameters(
        Names=["/aws/service/ami-amazon-linux-latest/al2023-ami-kernel-6.1-x86_64"]
    )
    ami = resp["Parameters"][0]["Value"]

    vpc = ec2.create_vpc(CidrBlock="172.28.7.0/24", InstanceTenancy="default")
    igw_id = ec2_client.create_internet_gateway()["InternetGateway"][
        "InternetGatewayId"
    ]
    vpc.attach_internet_gateway(InternetGatewayId=igw_id)
    route_table = vpc.create_route_table()
    route_table.create_route(DestinationCidrBlock="0.0.0.0/0", GatewayId=igw_id)
    security_group = ec2.create_security_group(
        GroupName=sg_name, Description="a", VpcId=vpc.id
    )

    ec2_client.authorize_security_group_ingress(
        GroupId=security_group.id,
        IpPermissions=[{"IpProtocol": "-1", "IpRanges": [{"CidrIp": "0.0.0.0/0"}]}],
    )
    subnet1 = vpc.create_subnet(
        CidrBlock="172.28.7.192/26", AvailabilityZone="us-east-1a"
    )
    subnet2 = vpc.create_subnet(
        CidrBlock="172.28.7.0/26", AvailabilityZone="us-east-1b"
    )

    route_table.associate_with_subnet(SubnetId=subnet1.id)
    route_table.associate_with_subnet(SubnetId=subnet2.id)

    load_balancer_arn = listener_arn = target_group_arn = instance_id1 = (
        instance_id2
    ) = None

    try:
        # Init-script that starts Apache
        init_script = """#!/bin/bash
sudo yum update -y
sudo yum install -y httpd
sudo systemctl start httpd
sudo systemctl enable httpd"""

        # Start EC2 instances first, to give them time to get ready
        instance_id1 = ec2.create_instances(
            ImageId=ami,
            MinCount=1,
            MaxCount=1,
            KeyName="test2",
            NetworkInterfaces=[
                {
                    "DeviceIndex": 0,
                    "SubnetId": subnet1.id,
                    "AssociatePublicIpAddress": True,
                    "Groups": [security_group.id],
                }
            ],
            UserData=init_script,
        )[0].id
        instance_id2 = ec2.create_instances(
            ImageId=ami,
            MinCount=1,
            MaxCount=1,
            KeyName="test2",
            NetworkInterfaces=[
                {
                    "DeviceIndex": 0,
                    "SubnetId": subnet2.id,
                    "AssociatePublicIpAddress": True,
                    "Groups": [security_group.id],
                }
            ],
            UserData=init_script,
        )[0].id

        waiter = ec2_client.get_waiter("instance_running")
        waiter.wait(InstanceIds=[instance_id1, instance_id2])

        load_balancer_arn = conn.create_load_balancer(
            Name=lb_name,
            Subnets=[subnet1.id, subnet2.id],
            SecurityGroups=[security_group.id],
            Scheme="internal",
            Tags=[{"Key": "key_name", "Value": "a_value"}],
        )["LoadBalancers"][0]["LoadBalancerArn"]
        waiter = conn.get_waiter("load_balancer_available")
        waiter.wait(LoadBalancerArns=[load_balancer_arn])

        target_group_arn = conn.create_target_group(
            Name=target_name,
            Protocol="HTTP",
            Port=80,
            VpcId=vpc.id,
            HealthCheckProtocol="HTTP",
            HealthCheckPort="80",
            HealthCheckPath="/",
            HealthCheckIntervalSeconds=5,
            HealthCheckTimeoutSeconds=3,
            HealthyThresholdCount=5,
            UnhealthyThresholdCount=2,
            # Apache returns a 403 (with text: 'It Works!') by default if no other pages are available
            Matcher={"HttpCode": "403"},
        )["TargetGroups"][0]["TargetGroupArn"]

        # No targets registered yet
        healths = conn.describe_target_health(TargetGroupArn=target_group_arn)[
            "TargetHealthDescriptions"
        ]
        assert len(healths) == 0

        listener_arn = conn.create_listener(
            LoadBalancerArn=load_balancer_arn,
            Protocol="HTTP",
            Port=80,
            DefaultActions=[
                {
                    "Type": "forward",
                    "TargetGroupArn": target_group_arn,
                }
            ],
        )["Listeners"][0]["ListenerArn"]

        # Register target
        # Verify that they are healthy
        conn.register_targets(
            TargetGroupArn=target_group_arn,
            Targets=[{"Id": instance_id1}, {"Id": instance_id2}],
        )
        waiter = conn.get_waiter("target_in_service")
        waiter.wait(TargetGroupArn=target_group_arn)

        healths = conn.describe_target_health(TargetGroupArn=target_group_arn)[
            "TargetHealthDescriptions"
        ]
        assert {
            "Target": {"Id": instance_id1, "Port": 80},
            "HealthCheckPort": "80",
            "TargetHealth": {"State": "healthy"},
        } in healths
        assert {
            "Target": {"Id": instance_id2, "Port": 80},
            "HealthCheckPort": "80",
            "TargetHealth": {"State": "healthy"},
        } in healths

        # De-register target
        # Verify that they no longer appear when describing health
        conn.deregister_targets(
            TargetGroupArn=target_group_arn,
            Targets=[{"Id": instance_id1}],
        )
        waiter = conn.get_waiter("target_deregistered")
        waiter.wait(TargetGroupArn=target_group_arn, Targets=[{"Id": instance_id1}])

        healths = conn.describe_target_health(TargetGroupArn=target_group_arn)[
            "TargetHealthDescriptions"
        ]
        assert len(healths) == 1
        assert {
            "Target": {"Id": instance_id2, "Port": 80},
            "HealthCheckPort": "80",
            "TargetHealth": {"State": "healthy"},
        } in healths

        # We can explicitly request it's health though
        healths = conn.describe_target_health(
            TargetGroupArn=target_group_arn, Targets=[{"Id": instance_id1, "Port": 80}]
        )["TargetHealthDescriptions"]
        assert len(healths) == 1
        assert {
            "Target": {"Id": instance_id1, "Port": 80},
            "HealthCheckPort": "80",
            "TargetHealth": {
                "State": "unused",
                "Reason": "Target.NotRegistered",
                "Description": "Target is not registered to the target group",
            },
        } in healths

        # Re-register it, and stop the instance
        conn.register_targets(
            TargetGroupArn=target_group_arn,
            Targets=[{"Id": instance_id1}],
        )
        waiter = conn.get_waiter("target_in_service")
        waiter.wait(TargetGroupArn=target_group_arn)

        healths = conn.describe_target_health(TargetGroupArn=target_group_arn)[
            "TargetHealthDescriptions"
        ]
        assert {
            "Target": {"Id": instance_id1, "Port": 80},
            "HealthCheckPort": "80",
            "TargetHealth": {"State": "healthy"},
        } in healths
        assert {
            "Target": {"Id": instance_id2, "Port": 80},
            "HealthCheckPort": "80",
            "TargetHealth": {"State": "healthy"},
        } in healths

        ec2_client.stop_instances(InstanceIds=[instance_id1])
        waiter = ec2_client.get_waiter("instance_stopped")
        waiter.wait(InstanceIds=[instance_id1])

        # Instance is marked as 'unused' when stopped
        states = set()
        while states != {"healthy", "unused"}:
            healths = conn.describe_target_health(TargetGroupArn=target_group_arn)[
                "TargetHealthDescriptions"
            ]
            states = set([h["TargetHealth"]["State"] for h in healths])

            if target_is_aws:
                sleep(5)
        assert {
            "Target": {"Id": instance_id2, "Port": 80},
            "HealthCheckPort": "80",
            "TargetHealth": {"State": "healthy"},
        } in healths
        assert {
            "Target": {"Id": instance_id1, "Port": 80},
            "HealthCheckPort": "80",
            "TargetHealth": {
                "State": "unused",
                "Reason": "Target.InvalidState",
                "Description": "Target is in the stopped state",
            },
        } in healths

        ec2_client.terminate_instances(InstanceIds=[instance_id1])
        waiter = ec2_client.get_waiter("instance_terminated")
        waiter.wait(InstanceIds=[instance_id1])

        # Instance is removed when terminated
        healths = conn.describe_target_health(TargetGroupArn=target_group_arn)[
            "TargetHealthDescriptions"
        ]
        while len(healths) != 1:
            healths = conn.describe_target_health(TargetGroupArn=target_group_arn)[
                "TargetHealthDescriptions"
            ]

            if target_is_aws:
                sleep(5)

        assert {
            "Target": {"Id": instance_id2, "Port": 80},
            "HealthCheckPort": "80",
            "TargetHealth": {"State": "healthy"},
        } in healths

        # Terminated instances are marked as 'draining' until eternity (well - atleast 10 minutes)
        healths = conn.describe_target_health(
            TargetGroupArn=target_group_arn, Targets=[{"Id": instance_id1, "Port": 80}]
        )["TargetHealthDescriptions"]
        assert healths == [
            {
                "Target": {"Id": instance_id1, "Port": 80},
                "HealthCheckPort": "80",
                "TargetHealth": {
                    "State": "draining",
                    "Reason": "Target.DeregistrationInProgress",
                    "Description": "Target deregistration is in progress",
                },
            }
        ]

    finally:
        try:  # to delete instance1 - just in case it failed to stop earlier
            ec2_client.stop_instances(InstanceIds=[instance_id1])
            waiter = ec2_client.get_waiter("instance_stopped")
            waiter.wait(InstanceIds=[instance_id1])
            ec2_client.terminate_instances(InstanceIds=[instance_id1])
            waiter = ec2_client.get_waiter("instance_terminated")
            waiter.wait(InstanceIds=[instance_id1])
        except:  # noqa: E722 Do not use bare except
            pass

        if instance_id2:
            ec2_client.stop_instances(InstanceIds=[instance_id2])
            waiter = ec2_client.get_waiter("instance_stopped")
            waiter.wait(InstanceIds=[instance_id2])
            ec2_client.terminate_instances(InstanceIds=[instance_id2])
            waiter = ec2_client.get_waiter("instance_terminated")
            waiter.wait(InstanceIds=[instance_id2])

        if listener_arn:
            conn.delete_listener(ListenerArn=listener_arn)

        if target_group_arn:
            conn.delete_target_group(TargetGroupArn=target_group_arn)

        if load_balancer_arn:
            conn.delete_load_balancer(LoadBalancerArn=load_balancer_arn)
            waiter = conn.get_waiter("load_balancers_deleted")
            waiter.wait(LoadBalancerArns=[load_balancer_arn])

    if target_is_aws:
        # Resources take a while to deregister
        # Wait, so we don't get DependencyViolations when deleting the SG
        sleep(30)

        security_group.delete()
        subnet1.delete()
        subnet2.delete()
        route_table.delete()
        ec2_client.detach_internet_gateway(InternetGatewayId=igw_id, VpcId=vpc.id)
        vpc.delete()
        ec2_client.delete_internet_gateway(InternetGatewayId=igw_id)


@elbv2_aws_verified()
@pytest.mark.aws_verified
def test_describe_unknown_targets(target_is_aws=False):
    conn = boto3.client("elbv2", region_name="us-east-1")
    ec2 = boto3.resource("ec2", region_name="us-east-1")
    ec2_client = boto3.client("ec2", region_name="us-east-1")

    lb_name = f"lb-{str(uuid4())[0:6]}"
    sg_name = f"sg{str(uuid4())[0:6]}"
    target_name = f"target-{str(uuid4())[0:6]}"

    vpc = ec2.create_vpc(CidrBlock="172.28.7.0/24", InstanceTenancy="default")
    security_group = ec2.create_security_group(
        GroupName=sg_name, Description="a", VpcId=vpc.id
    )

    ec2_client.authorize_security_group_ingress(
        GroupId=security_group.id,
        IpPermissions=[{"IpProtocol": "-1", "IpRanges": [{"CidrIp": "0.0.0.0/0"}]}],
    )
    subnet1 = vpc.create_subnet(
        CidrBlock="172.28.7.192/26", AvailabilityZone="us-east-1a"
    )
    subnet2 = vpc.create_subnet(
        CidrBlock="172.28.7.0/26", AvailabilityZone="us-east-1b"
    )

    load_balancer_arn = listener_arn = target_group_arn = None
    instance_id = "i-12345678901234567"

    try:
        load_balancer_arn = conn.create_load_balancer(
            Name=lb_name,
            Subnets=[subnet1.id, subnet2.id],
            SecurityGroups=[security_group.id],
            Scheme="internal",
            Tags=[{"Key": "key_name", "Value": "a_value"}],
        )["LoadBalancers"][0]["LoadBalancerArn"]
        waiter = conn.get_waiter("load_balancer_available")
        waiter.wait(LoadBalancerArns=[load_balancer_arn])

        target_group_arn = conn.create_target_group(
            Name=target_name,
            Protocol="HTTP",
            Port=80,
            VpcId=vpc.id,
            HealthCheckProtocol="HTTP",
            HealthCheckPort="80",
            HealthCheckPath="/",
            HealthCheckIntervalSeconds=5,
            HealthCheckTimeoutSeconds=3,
            HealthyThresholdCount=5,
            UnhealthyThresholdCount=2,
            # Apache returns a 403 (with text: 'It Works!') by default if no other pages are available
            Matcher={"HttpCode": "403"},
        )["TargetGroups"][0]["TargetGroupArn"]

        listener_arn = conn.create_listener(
            LoadBalancerArn=load_balancer_arn,
            Protocol="HTTP",
            Port=80,
            DefaultActions=[
                {
                    "Type": "forward",
                    "TargetGroupArn": target_group_arn,
                }
            ],
        )["Listeners"][0]["ListenerArn"]

        healths = conn.describe_target_health(
            TargetGroupArn=target_group_arn, Targets=[{"Id": instance_id}]
        )["TargetHealthDescriptions"]
        assert healths == [
            {
                "Target": {"Id": instance_id, "Port": 80},
                "HealthCheckPort": "80",
                "TargetHealth": {
                    "State": "unused",
                    "Reason": "Target.NotRegistered",
                    "Description": "Target is not registered to the target group",
                },
            }
        ]

    finally:
        if listener_arn:
            conn.delete_listener(ListenerArn=listener_arn)

        if target_group_arn:
            conn.delete_target_group(TargetGroupArn=target_group_arn)

        if load_balancer_arn:
            conn.delete_load_balancer(LoadBalancerArn=load_balancer_arn)
            waiter = conn.get_waiter("load_balancers_deleted")
            waiter.wait(LoadBalancerArns=[load_balancer_arn])

    if target_is_aws:
        # Resources take a while to deregister
        # Wait, so we don't get DependencyViolations when deleting the SG
        sleep(30)

        security_group.delete()
        subnet1.delete()
        subnet2.delete()
        vpc.delete()



"""Exceptions raised by the opensearchserverless service."""

from moto.core.exceptions import JsonRESTError


class BackupClientError(JsonRESTError):
    code = 400


class ConflictException(BackupClientError):
    def __init__(self, msg: str):
        super().__init__("ConflictException", f"{msg}")


class ValidationException(BackupClientError):
    def __init__(self, msg: str):
        super().__init__("ValidationException", f"{msg}")


class ResourceNotFoundException(BackupClientError):
    def __init__(self, msg: str):
        super().__init__("ResourceNotFoundException", f"{msg}")



import boto3

from moto import mock_aws


@mock_aws
def test_list_rule_groups():
    client = boto3.client("wafv2", region_name="us-east-2")
    resp = client.list_rule_groups(Scope="CLOUDFRONT")
    assert resp["RuleGroups"] == []



import boto3

from moto import mock_aws


@mock_aws
def test_describe_maintenance_window():
    ssm = boto3.client("ssm", region_name="us-east-1")

    resp = ssm.describe_maintenance_windows()
    assert resp["WindowIdentities"] == []

    resp = ssm.describe_maintenance_windows(
        Filters=[{"Key": "Name", "Values": ["fake-maintenance-window-name"]}]
    )
    assert resp["WindowIdentities"] == []


@mock_aws
def test_create_maintenance_windows_simple():
    ssm = boto3.client("ssm", region_name="us-east-1")

    resp = ssm.create_maintenance_window(
        Name="simple-window",
        Schedule="cron(15 12 * * ? *)",
        Duration=2,
        Cutoff=1,
        AllowUnassociatedTargets=False,
    )
    assert "WindowId" in resp
    _id = resp["WindowId"]  # mw-01d6bbfdf6af2c39a

    resp = ssm.describe_maintenance_windows()
    assert len(resp["WindowIdentities"]) == 1

    my_window = resp["WindowIdentities"][0]
    assert my_window["WindowId"] == _id
    assert my_window["Name"] == "simple-window"
    assert my_window["Enabled"] is True
    assert my_window["Duration"] == 2
    assert my_window["Cutoff"] == 1
    assert my_window["Schedule"] == "cron(15 12 * * ? *)"
    # assert "NextExecutionTime" in my_window
    assert "Description" not in my_window
    assert "ScheduleTimezone" not in my_window
    assert "ScheduleOffset" not in my_window
    assert "EndDate" not in my_window
    assert "StartDate" not in my_window


@mock_aws
def test_create_maintenance_windows_advanced():
    ssm = boto3.client("ssm", region_name="us-east-1")

    resp = ssm.create_maintenance_window(
        Name="simple-window",
        Description="French windows are just too fancy",
        Schedule="cron(15 12 * * ? *)",
        ScheduleTimezone="Europe/London",
        ScheduleOffset=1,
        Duration=5,
        Cutoff=4,
        AllowUnassociatedTargets=False,
        StartDate="2021-11-01",
        EndDate="2021-12-31",
    )
    assert "WindowId" in resp
    _id = resp["WindowId"]  # mw-01d6bbfdf6af2c39a

    resp = ssm.describe_maintenance_windows()
    assert len(resp["WindowIdentities"]) == 1

    my_window = resp["WindowIdentities"][0]
    assert my_window["WindowId"] == _id
    assert my_window["Name"] == "simple-window"
    assert my_window["Enabled"] is True
    assert my_window["Duration"] == 5
    assert my_window["Cutoff"] == 4
    assert my_window["Schedule"] == "cron(15 12 * * ? *)"
    # assert "NextExecutionTime" in my_window
    assert my_window["Description"] == "French windows are just too fancy"
    assert my_window["ScheduleTimezone"] == "Europe/London"
    assert my_window["ScheduleOffset"] == 1
    assert my_window["StartDate"] == "2021-11-01"
    assert my_window["EndDate"] == "2021-12-31"


@mock_aws
def test_get_maintenance_windows():
    ssm = boto3.client("ssm", region_name="us-east-1")

    resp = ssm.create_maintenance_window(
        Name="my-window",
        Schedule="cron(15 12 * * ? *)",
        Duration=2,
        Cutoff=1,
        AllowUnassociatedTargets=False,
    )
    assert "WindowId" in resp
    _id = resp["WindowId"]  # mw-01d6bbfdf6af2c39a

    my_window = ssm.get_maintenance_window(WindowId=_id)
    assert my_window["WindowId"] == _id
    assert my_window["Name"] == "my-window"
    assert my_window["Enabled"] is True
    assert my_window["Duration"] == 2
    assert my_window["Cutoff"] == 1
    assert my_window["Schedule"] == "cron(15 12 * * ? *)"
    # assert "NextExecutionTime" in my_window
    assert "Description" not in my_window
    assert "ScheduleTimezone" not in my_window
    assert "ScheduleOffset" not in my_window
    assert "EndDate" not in my_window
    assert "StartDate" not in my_window


@mock_aws
def test_describe_maintenance_windows():
    ssm = boto3.client("ssm", region_name="us-east-1")

    for idx in range(0, 4):
        ssm.create_maintenance_window(
            Name=f"window_{idx}",
            Schedule="cron(15 12 * * ? *)",
            Duration=2,
            Cutoff=1,
            AllowUnassociatedTargets=False,
        )

    resp = ssm.describe_maintenance_windows()
    assert len(resp["WindowIdentities"]) == 4

    resp = ssm.describe_maintenance_windows(
        Filters=[{"Key": "Name", "Values": ["window_0", "window_2"]}]
    )
    assert len(resp["WindowIdentities"]) == 2


@mock_aws
def test_delete_maintenance_windows():
    ssm = boto3.client("ssm", region_name="us-east-1")

    resp = ssm.create_maintenance_window(
        Name="simple-window",
        Schedule="cron(15 12 * * ? *)",
        Duration=2,
        Cutoff=1,
        AllowUnassociatedTargets=False,
    )

    ssm.delete_maintenance_window(WindowId=resp["WindowId"])

    resp = ssm.describe_maintenance_windows()
    assert resp["WindowIdentities"] == []


@mock_aws
def test_tags():
    ssm = boto3.client("ssm", region_name="us-east-1")

    # create without & list
    mw_id = ssm.create_maintenance_window(
        Name="simple-window",
        Schedule="cron(15 12 * * ? *)",
        Duration=2,
        Cutoff=1,
        AllowUnassociatedTargets=False,
    )["WindowId"]

    # add & list
    ssm.add_tags_to_resource(
        ResourceType="MaintenanceWindow",
        ResourceId=mw_id,
        Tags=[{"Key": "k1", "Value": "v1"}],
    )
    tags = ssm.list_tags_for_resource(
        ResourceType="MaintenanceWindow", ResourceId=mw_id
    )["TagList"]
    assert tags == [{"Key": "k1", "Value": "v1"}]

    # create & list
    mw_id = ssm.create_maintenance_window(
        Name="simple-window",
        Schedule="cron(15 12 * * ? *)",
        Duration=2,
        Cutoff=1,
        AllowUnassociatedTargets=False,
        Tags=[{"Key": "k2", "Value": "v2"}],
    )["WindowId"]
    tags = ssm.list_tags_for_resource(
        ResourceType="MaintenanceWindow", ResourceId=mw_id
    )["TagList"]
    assert tags == [{"Key": "k2", "Value": "v2"}]

    # add more & list
    ssm.add_tags_to_resource(
        ResourceType="MaintenanceWindow",
        ResourceId=mw_id,
        Tags=[{"Key": "k3", "Value": "v3"}],
    )
    tags = ssm.list_tags_for_resource(
        ResourceType="MaintenanceWindow", ResourceId=mw_id
    )["TagList"]
    assert tags == [{"Key": "k2", "Value": "v2"}, {"Key": "k3", "Value": "v3"}]

    # remove & list
    ssm.remove_tags_from_resource(
        ResourceType="MaintenanceWindow", ResourceId=mw_id, TagKeys=["k3"]
    )
    tags = ssm.list_tags_for_resource(
        ResourceType="MaintenanceWindow", ResourceId=mw_id
    )["TagList"]
    assert tags == [{"Key": "k2", "Value": "v2"}]


@mock_aws
def test_register_maintenance_window_target():
    ssm = boto3.client("ssm", region_name="us-east-1")

    resp = ssm.create_maintenance_window(
        Name="simple-window",
        Schedule="cron(15 12 * * ? *)",
        Duration=2,
        Cutoff=1,
        AllowUnassociatedTargets=False,
    )
    window_id = resp["WindowId"]

    resp = ssm.register_target_with_maintenance_window(
        WindowId=window_id,
        ResourceType="INSTANCE",
        Targets=[{"Key": "tag:Name", "Values": ["my-instance"]}],
    )
    assert "WindowTargetId" in resp
    _id = resp["WindowTargetId"]

    resp = ssm.describe_maintenance_window_targets(
        WindowId=window_id,
    )
    assert len(resp["Targets"]) == 1
    assert resp["Targets"][0]["ResourceType"] == "INSTANCE"
    assert resp["Targets"][0]["WindowTargetId"] == _id
    assert resp["Targets"][0]["Targets"][0]["Key"] == "tag:Name"
    assert resp["Targets"][0]["Targets"][0]["Values"] == ["my-instance"]


@mock_aws
def test_deregister_target_from_maintenance_window():
    ssm = boto3.client("ssm", region_name="us-east-1")

    resp = ssm.create_maintenance_window(
        Name="simple-window",
        Schedule="cron(15 12 * * ? *)",
        Duration=2,
        Cutoff=1,
        AllowUnassociatedTargets=False,
    )
    window_id = resp["WindowId"]

    resp = ssm.register_target_with_maintenance_window(
        WindowId=window_id,
        ResourceType="INSTANCE",
        Targets=[{"Key": "tag:Name", "Values": ["my-instance"]}],
    )
    _id = resp["WindowTargetId"]

    ssm.deregister_target_from_maintenance_window(
        WindowId=window_id,
        WindowTargetId=_id,
    )

    resp = ssm.describe_maintenance_window_targets(
        WindowId=window_id,
    )
    assert len(resp["Targets"]) == 0


@mock_aws
def test_describe_maintenance_window_with_no_task_or_targets():
    ssm = boto3.client("ssm", region_name="us-east-1")

    resp = ssm.create_maintenance_window(
        Name="simple-window",
        Schedule="cron(15 12 * * ? *)",
        Duration=2,
        Cutoff=1,
        AllowUnassociatedTargets=False,
    )
    window_id = resp["WindowId"]

    resp = ssm.describe_maintenance_window_tasks(
        WindowId=window_id,
    )
    assert len(resp["Tasks"]) == 0

    resp = ssm.describe_maintenance_window_targets(
        WindowId=window_id,
    )
    assert len(resp["Targets"]) == 0


@mock_aws
def test_register_maintenance_window_task():
    ssm = boto3.client("ssm", region_name="us-east-1")

    resp = ssm.create_maintenance_window(
        Name="simple-window",
        Schedule="cron(15 12 * * ? *)",
        Duration=2,
        Cutoff=1,
        AllowUnassociatedTargets=False,
    )
    window_id = resp["WindowId"]

    resp = ssm.register_target_with_maintenance_window(
        WindowId=window_id,
        ResourceType="INSTANCE",
        Targets=[{"Key": "tag:Name", "Values": ["my-instance"]}],
    )
    window_target_id = resp["WindowTargetId"]

    resp = ssm.register_task_with_maintenance_window(
        WindowId=window_id,
        Targets=[{"Key": "WindowTargetIds", "Values": [window_target_id]}],
        TaskArn="AWS-RunShellScript",
        TaskType="RUN_COMMAND",
        MaxConcurrency="1",
        MaxErrors="1",
    )

    assert "WindowTaskId" in resp
    _id = resp["WindowTaskId"]

    resp = ssm.describe_maintenance_window_tasks(
        WindowId=window_id,
    )
    assert len(resp["Tasks"]) == 1
    assert resp["Tasks"][0]["WindowTaskId"] == _id
    assert resp["Tasks"][0]["WindowId"] == window_id
    assert resp["Tasks"][0]["TaskArn"] == "AWS-RunShellScript"
    assert resp["Tasks"][0]["MaxConcurrency"] == "1"
    assert resp["Tasks"][0]["MaxErrors"] == "1"


@mock_aws
def test_deregister_maintenance_window_task():
    ssm = boto3.client("ssm", region_name="us-east-1")

    resp = ssm.create_maintenance_window(
        Name="simple-window",
        Schedule="cron(15 12 * * ? *)",
        Duration=2,
        Cutoff=1,
        AllowUnassociatedTargets=False,
    )
    window_id = resp["WindowId"]

    resp = ssm.register_target_with_maintenance_window(
        WindowId=window_id,
        ResourceType="INSTANCE",
        Targets=[{"Key": "tag:Name", "Values": ["my-instance"]}],
    )
    window_target_id = resp["WindowTargetId"]

    resp = ssm.register_task_with_maintenance_window(
        WindowId=window_id,
        Targets=[{"Key": "WindowTargetIds", "Values": [window_target_id]}],
        TaskArn="AWS-RunShellScript",
        TaskType="RUN_COMMAND",
        MaxConcurrency="1",
        MaxErrors="1",
    )
    window_task_id = resp["WindowTaskId"]

    ssm.deregister_task_from_maintenance_window(
        WindowId=window_id,
        WindowTaskId=window_task_id,
    )

    resp = ssm.describe_maintenance_window_tasks(
        WindowId=window_id,
    )
    assert len(resp["Tasks"]) == 0



from moto.ec2.utils import add_tag_specification

from ._base_response import EC2BaseResponse


class EgressOnlyInternetGateway(EC2BaseResponse):
    def create_egress_only_internet_gateway(self) -> str:
        vpc_id = self._get_param("VpcId")
        tag_param = self._get_multi_param("TagSpecification")
        tags = add_tag_specification(tag_param)

        egress_only_igw = self.ec2_backend.create_egress_only_internet_gateway(
            vpc_id=vpc_id, tags=tags
        )
        template = self.response_template(CREATE_EGRESS_ONLY_IGW_RESPONSE)
        return template.render(egress_only_igw=egress_only_igw)

    def describe_egress_only_internet_gateways(self) -> str:
        egress_only_igw_ids = self._get_multi_param("EgressOnlyInternetGatewayId")
        egress_only_igws = self.ec2_backend.describe_egress_only_internet_gateways(
            egress_only_igw_ids
        )
        template = self.response_template(DESCRIBE_EGRESS_ONLY_IGW_RESPONSE)
        return template.render(egress_only_igws=egress_only_igws)

    def delete_egress_only_internet_gateway(self) -> str:
        egress_only_igw_id = self._get_param("EgressOnlyInternetGatewayId")
        self.ec2_backend.delete_egress_only_internet_gateway(
            gateway_id=egress_only_igw_id
        )
        template = self.response_template(DELETE_EGRESS_ONLY_IGW_RESPONSE)
        return template.render()


CREATE_EGRESS_ONLY_IGW_RESPONSE = """<CreateEgressOnlyInternetGatewayResponse xmlns="http://ec2.amazonaws.com/doc/2016-11-15/">
    <requestId>c617595f-6c29-4a00-a941-example</requestId>
    <egressOnlyInternetGateway>
        <attachmentSet>
            <item>
                <state>{{ egress_only_igw.state }}</state>
                <vpcId>{{ egress_only_igw.vpc_id }}</vpcId>
            </item>
        </attachmentSet>
        <egressOnlyInternetGatewayId>{{ egress_only_igw.id }}</egressOnlyInternetGatewayId>
        <tagSet>
        {% for tag in egress_only_igw.get_tags() %}
            <item>
                <key>{{ tag.key }}</key>
                <value>{{ tag.value }}</value>
            </item>
        {% endfor %}
        </tagSet>
    </egressOnlyInternetGateway>
</CreateEgressOnlyInternetGatewayResponse>
"""

DESCRIBE_EGRESS_ONLY_IGW_RESPONSE = """<DescribeEgressOnlyInternetGatewaysResponse xmlns="http://ec2.amazonaws.com/doc/2016-11-15/">
    <requestId>ec441b4c-357f-4483-b4a7-example</requestId>
    <egressOnlyInternetGatewaySet>
        {% for egress_only_igw in egress_only_igws %}
        <item>
            <attachmentSet>
                <item>
                    <state>{{ egress_only_igw.state }}</state>
                    <vpcId>{{ egress_only_igw.vpc_id }}</vpcId>
                </item>
            </attachmentSet>
            <egressOnlyInternetGatewayId>{{ egress_only_igw.id }}</egressOnlyInternetGatewayId>
            <tagSet>
            {% for tag in egress_only_igw.get_tags() %}
                <item>
                    <key>{{ tag.key }}</key>
                    <value>{{ tag.value }}</value>
                </item>
            {% endfor %}
            </tagSet>
        </item>
        {% endfor %}
    </egressOnlyInternetGatewaySet>
</DescribeEgressOnlyInternetGatewaysResponse>"""

DELETE_EGRESS_ONLY_IGW_RESPONSE = """<DeleteEgressOnlyInternetGateway xmlns="http://ec2.amazonaws.com/doc/2016-11-15/">
  <requestId>59dbff89-35bd-4eac-99ed-be587EXAMPLE</requestId>
  <returnCode>true</returnCode>
</DeleteEgressOnlyInternetGateway>"""



from .responses import IoTDataPlaneResponse

url_bases = [
    r"https?://data\.iot\.(.+)\.amazonaws.com",
    r"https?://data-ats\.iot\.(.+)\.amazonaws.com",
]


response = IoTDataPlaneResponse()


url_paths = {
    "{0}/.*$": response.dispatch,
    #
    # (Flask) Paths for :class:`moto.core.models.ServerModeMockAWS`
    #
    "{0}/<path:route>$": response.dispatch,
}



from moto.core.exceptions import JsonRESTError


class MediaPackageClientError(JsonRESTError):
    code = 400


# AWS service exceptions are caught with the underlying botocore exception, ClientError
class ClientError(MediaPackageClientError):
    def __init__(self, error: str, message: str):
        super().__init__(error, message)



import boto3
import pytest
from botocore.exceptions import ClientError  # Boto3 will always throw this exception

from moto import mock_aws

region = "eu-west-1"


def _create_channel_config(**kwargs):
    channel_id = kwargs.get("id", "channel-id")
    description = kwargs.get("description", "Awesome channel!")
    tags = kwargs.get("tags", {"Customer": "moto"})
    channel_config = dict(Description=description, Id=channel_id, Tags=tags)
    return channel_config


def _create_origin_endpoint_config(**kwargs):
    authorization = kwargs.get(
        "authorization",
        {"CdnIdentifierSecret": "cdn-id-secret", "SecretsRoleArn": "secrets-arn"},
    )
    channel_id = kwargs.get("channel_id", "channel-id")
    cmaf_package = kwargs.get("cmafpackage", {"HlsManifests": []})
    dash_package = kwargs.get("dash_package", {"AdTriggers": []})
    description = kwargs.get("description", "channel-description")
    hls_package = kwargs.get("hls_package", {"AdMarkers": "NONE"})
    endpoint_id = kwargs.get("id", "origin-endpoint-id")
    manifest_name = kwargs.get("manifest_name", "manifest-name")
    mss_package = kwargs.get("mss_package", {"ManifestWindowSeconds": 1})
    origination = kwargs.get("origination", "ALLOW")
    startover_window_seconds = kwargs.get("startover_window_seconds", 1)
    tags = kwargs.get("tags", {"Customer": "moto"})
    time_delay_seconds = kwargs.get("time_delay_seconds", 1)
    whitelist = kwargs.get("whitelist", ["whitelist"])
    origin_endpoint_config = dict(
        Authorization=authorization,
        ChannelId=channel_id,
        CmafPackage=cmaf_package,
        DashPackage=dash_package,
        Description=description,
        HlsPackage=hls_package,
        Id=endpoint_id,
        ManifestName=manifest_name,
        MssPackage=mss_package,
        Origination=origination,
        StartoverWindowSeconds=startover_window_seconds,
        Tags=tags,
        TimeDelaySeconds=time_delay_seconds,
        Whitelist=whitelist,
    )
    return origin_endpoint_config


@mock_aws
def test_create_channel_succeeds():
    client = boto3.client("mediapackage", region_name=region)
    channel_config = _create_channel_config()

    channel = client.create_channel(**channel_config)
    assert channel["ResponseMetadata"]["HTTPStatusCode"] == 200
    assert channel["Arn"] == f"arn:aws:mediapackage:channel:{channel['Id']}"
    assert channel["Description"] == "Awesome channel!"
    assert channel["Id"] == "channel-id"
    assert channel["Tags"]["Customer"] == "moto"


@mock_aws
def test_describe_channel_succeeds():
    client = boto3.client("mediapackage", region_name=region)
    channel_config = _create_channel_config()

    channel_id = client.create_channel(**channel_config)["Id"]
    channel = client.describe_channel(Id=channel_id)
    assert channel["Arn"] == f"arn:aws:mediapackage:channel:{channel['Id']}"
    assert channel["Description"] == channel_config["Description"]
    assert channel["Tags"]["Customer"] == "moto"


@mock_aws
def test_describe_unknown_channel_throws_error():
    client = boto3.client("mediapackage", region_name=region)
    channel_id = "unknown-channel"
    with pytest.raises(ClientError) as err:
        client.describe_channel(Id=channel_id)
    err = err.value.response["Error"]
    assert err["Code"] == "NotFoundException"
    assert err["Message"] == f"channel with id={channel_id} not found"


@mock_aws
def test_delete_unknown_channel_throws_error():
    client = boto3.client("mediapackage", region_name=region)
    channel_id = "unknown-channel"
    with pytest.raises(ClientError) as err:
        client.delete_channel(Id=channel_id)
    err = err.value.response["Error"]
    assert err["Code"] == "NotFoundException"
    assert err["Message"] == f"channel with id={channel_id} not found"


@mock_aws
def test_delete_channel_successfully_deletes():
    client = boto3.client("mediapackage", region_name=region)
    channel_config = _create_channel_config()
    create_response = client.create_channel(**channel_config)
    # Before deletion
    list_response = client.list_channels()
    channels_list = list_response["Channels"]
    client.delete_channel(Id=create_response["Id"])
    # After deletion
    post_deletion_list_response = client.list_channels()
    post_deletion_channels_list = post_deletion_list_response["Channels"]
    assert len(post_deletion_channels_list) == len(channels_list) - 1


@mock_aws
def test_list_channels_succeds():
    client = boto3.client("mediapackage", region_name=region)
    channel_config = _create_channel_config()

    client.create_channel(**channel_config)
    channels_list = client.list_channels()["Channels"]
    assert len(channels_list) == 1

    channel = channels_list[0]
    assert channel["Arn"] == f"arn:aws:mediapackage:channel:{channel['Id']}"
    assert channel["Description"] == channel_config["Description"]
    assert channel["Tags"]["Customer"] == "moto"


@mock_aws
def test_create_origin_endpoint_succeeds():
    client = boto3.client("mediapackage", region_name=region)
    origin_endpoint_config = _create_origin_endpoint_config()

    endpoint = client.create_origin_endpoint(**origin_endpoint_config)
    assert endpoint["ResponseMetadata"]["HTTPStatusCode"] == 200
    assert endpoint["Arn"] == f"arn:aws:mediapackage:origin_endpoint:{endpoint['Id']}"
    assert endpoint["ChannelId"] == origin_endpoint_config["ChannelId"]
    assert endpoint["Description"] == origin_endpoint_config["Description"]
    assert endpoint["HlsPackage"] == origin_endpoint_config["HlsPackage"]
    assert endpoint["Origination"] == "ALLOW"


@mock_aws
def test_describe_origin_endpoint_succeeds():
    client = boto3.client("mediapackage", region_name=region)
    origin_endpoint_config = _create_origin_endpoint_config()

    endpoint_id = client.create_origin_endpoint(**origin_endpoint_config)["Id"]
    endpoint = client.describe_origin_endpoint(Id=endpoint_id)
    assert endpoint["ResponseMetadata"]["HTTPStatusCode"] == 200
    assert endpoint["Arn"] == f"arn:aws:mediapackage:origin_endpoint:{endpoint['Id']}"
    assert endpoint["ChannelId"] == origin_endpoint_config["ChannelId"]
    assert endpoint["Description"] == origin_endpoint_config["Description"]
    assert endpoint["HlsPackage"] == origin_endpoint_config["HlsPackage"]
    assert endpoint["Origination"] == "ALLOW"
    assert (
        endpoint["Url"]
        == f"https://origin-endpoint.mediapackage.{region}.amazonaws.com/{endpoint['Id']}"
    )


@mock_aws
def test_describe_unknown_origin_endpoint_throws_error():
    client = boto3.client("mediapackage", region_name=region)
    channel_id = "unknown-channel"
    with pytest.raises(ClientError) as err:
        client.describe_origin_endpoint(Id=channel_id)
    err = err.value.response["Error"]
    assert err["Code"] == "NotFoundException"
    assert err["Message"] == f"origin endpoint with id={channel_id} not found"


@mock_aws
def test_delete_origin_endpoint_succeeds():
    client = boto3.client("mediapackage", region_name=region)
    origin_endpoint_config = _create_origin_endpoint_config()
    create_response = client.create_origin_endpoint(**origin_endpoint_config)
    list_response = client.list_origin_endpoints()
    # Before deletion
    origin_endpoints_list = list_response["OriginEndpoints"]
    client.delete_origin_endpoint(Id=create_response["Id"])
    # After deletion
    post_deletion_list_response = client.list_origin_endpoints()
    post_deletion_origin_endpoints_list = post_deletion_list_response["OriginEndpoints"]
    assert len(post_deletion_origin_endpoints_list) == len(origin_endpoints_list) - 1


@mock_aws
def test_delete_unknown_origin_endpoint_throws_error():
    client = boto3.client("mediapackage", region_name=region)
    channel_id = "unknown-channel"
    with pytest.raises(ClientError) as err:
        client.delete_origin_endpoint(Id=channel_id)
    err = err.value.response["Error"]
    assert err["Code"] == "NotFoundException"
    assert err["Message"] == f"origin endpoint with id={channel_id} not found"


@mock_aws
def test_update_origin_endpoint_succeeds():
    client = boto3.client("mediapackage", region_name=region)
    origin_endpoint_config = _create_origin_endpoint_config()
    endpoint_id = client.create_origin_endpoint(**origin_endpoint_config)["Id"]

    endpoint = client.update_origin_endpoint(
        Id=endpoint_id,
        Description="updated-channel-description",
        ManifestName="updated-manifest-name",
    )
    assert endpoint["Description"] == "updated-channel-description"
    assert endpoint["ManifestName"] == "updated-manifest-name"


@mock_aws
def test_update_unknown_origin_endpoint_throws_error():
    client = boto3.client("mediapackage", region_name=region)
    channel_id = "unknown-channel"
    with pytest.raises(ClientError) as err:
        client.update_origin_endpoint(
            Id=channel_id,
            Description="updated-channel-description",
            ManifestName="updated-manifest-name",
        )
    err = err.value.response["Error"]
    assert err["Code"] == "NotFoundException"
    assert err["Message"] == f"origin endpoint with id={channel_id} not found"


@mock_aws
def test_list_origin_endpoint_succeeds():
    client = boto3.client("mediapackage", region_name=region)
    origin_endpoint_config = _create_origin_endpoint_config()

    client.create_origin_endpoint(**origin_endpoint_config)
    origin_endpoints_list = client.list_origin_endpoints()["OriginEndpoints"]
    assert len(origin_endpoints_list) == 1

    endpoint = origin_endpoints_list[0]
    assert endpoint["Arn"] == f"arn:aws:mediapackage:origin_endpoint:{endpoint['Id']}"
    assert endpoint["ChannelId"] == origin_endpoint_config["ChannelId"]
    assert endpoint["Description"] == origin_endpoint_config["Description"]
    assert endpoint["HlsPackage"] == origin_endpoint_config["HlsPackage"]
    assert endpoint["Origination"] == "ALLOW"



"""Handles incoming backup requests, invokes methods, returns responses."""

import json
from urllib.parse import unquote

from moto.core.responses import BaseResponse

from .models import BackupBackend, backup_backends


class BackupResponse(BaseResponse):
    """Handler for Backup requests and responses."""

    def __init__(self) -> None:
        super().__init__(service_name="backup")

    @property
    def backup_backend(self) -> BackupBackend:
        """Return backend instance specific for this region."""
        return backup_backends[self.current_account][self.region]

    def create_backup_plan(self) -> str:
        params = json.loads(self.body)
        backup_plan = params.get("BackupPlan")
        backup_plan_tags = params.get("BackupPlanTags")
        creator_request_id = params.get("CreatorRequestId")
        plan = self.backup_backend.create_backup_plan(
            backup_plan=backup_plan,
            backup_plan_tags=backup_plan_tags,
            creator_request_id=creator_request_id,
        )
        return json.dumps(dict(plan.to_dict()))

    def get_backup_plan(self) -> str:
        params = self._get_params()
        backup_plan_id = self.path.split("/")[-2]
        version_id = params.get("versionId")
        plan = self.backup_backend.get_backup_plan(
            backup_plan_id=backup_plan_id, version_id=version_id
        )
        return json.dumps(dict(plan.to_get_dict()))

    def delete_backup_plan(self) -> str:
        backup_plan_id = self.path.split("/")[-1]
        (
            backup_plan_id,
            backup_plan_arn,
            deletion_date,
            version_id,
        ) = self.backup_backend.delete_backup_plan(
            backup_plan_id=backup_plan_id,
        )
        return json.dumps(
            dict(
                BackupPlanId=backup_plan_id,
                BackupPlanArn=backup_plan_arn,
                DeletionDate=deletion_date,
                VersionId=version_id,
            )
        )

    def list_backup_plans(self) -> str:
        params = self._get_params()
        include_deleted = params.get("includeDeleted")
        backup_plans_list = self.backup_backend.list_backup_plans(
            include_deleted=include_deleted
        )
        return json.dumps(
            dict(BackupPlansList=[p.to_list_dict() for p in backup_plans_list])
        )

    def create_backup_vault(self) -> str:
        params = json.loads(self.body)
        backup_vault_name = self.path.split("/")[-1]
        backup_vault_tags = params.get("BackupVaultTags")
        encryption_key_arn = params.get("EncryptionKeyArn")
        creator_request_id = params.get("CreatorRequestId")
        backup_vault = self.backup_backend.create_backup_vault(
            backup_vault_name=backup_vault_name,
            backup_vault_tags=backup_vault_tags,
            encryption_key_arn=encryption_key_arn,
            creator_request_id=creator_request_id,
        )
        return json.dumps(dict(backup_vault.to_dict()))

    def list_backup_vaults(self) -> str:
        backup_vault_list = self.backup_backend.list_backup_vaults()
        return json.dumps(
            dict(BackupVaultList=[v.to_list_dict() for v in backup_vault_list])
        )

    def list_tags(self) -> str:
        resource_arn = unquote(self.path.split("/")[-2])
        tags = self.backup_backend.list_tags(
            resource_arn=resource_arn,
        )
        return json.dumps(dict(Tags=tags))

    def tag_resource(self) -> str:
        params = json.loads(self.body)
        resource_arn = unquote(self.path.split("/")[-1])
        tags = params.get("Tags")
        self.backup_backend.tag_resource(
            resource_arn=resource_arn,
            tags=tags,
        )
        return "{}"

    def untag_resource(self) -> str:
        params = json.loads(self.body)
        resource_arn = unquote(self.path.split("/")[-1])
        tag_key_list = params.get("TagKeyList")
        self.backup_backend.untag_resource(
            resource_arn=resource_arn,
            tag_key_list=tag_key_list,
        )
        return "{}"



from collections import OrderedDict
from typing import Any, Dict, List

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.utilities.utils import get_partition

from .exceptions import ClientError


class Channel(BaseModel):
    def __init__(self, **kwargs: Any):
        self.arn = kwargs.get("arn")
        self.channel_id = kwargs.get("channel_id")
        self.description = kwargs.get("description")
        self.tags = kwargs.get("tags")

    def to_dict(self) -> Dict[str, Any]:
        return {
            "arn": self.arn,
            "id": self.channel_id,
            "description": self.description,
            "tags": self.tags,
        }


class OriginEndpoint(BaseModel):
    def __init__(self, **kwargs: Any):
        self.arn = kwargs.get("arn")
        self.authorization = kwargs.get("authorization")
        self.channel_id = kwargs.get("channel_id")
        self.cmaf_package = kwargs.get("cmaf_package")
        self.dash_package = kwargs.get("dash_package")
        self.description = kwargs.get("description")
        self.hls_package = kwargs.get("hls_package")
        self.id = kwargs.get("endpoint_id")
        self.manifest_name = kwargs.get("manifest_name")
        self.mss_package = kwargs.get("mss_package")
        self.origination = kwargs.get("origination")
        self.startover_window_seconds = kwargs.get("startover_window_seconds")
        self.tags = kwargs.get("tags")
        self.time_delay_seconds = kwargs.get("time_delay_seconds")
        self.url = kwargs.get("url")
        self.whitelist = kwargs.get("whitelist")

    def to_dict(self) -> Dict[str, Any]:
        return {
            "arn": self.arn,
            "authorization": self.authorization,
            "channelId": self.channel_id,
            "cmafPackage": self.cmaf_package,
            "dashPackage": self.dash_package,
            "description": self.description,
            "hlsPackage": self.hls_package,
            "id": self.id,
            "manifestName": self.manifest_name,
            "mssPackage": self.mss_package,
            "origination": self.origination,
            "startoverWindowSeconds": self.startover_window_seconds,
            "tags": self.tags,
            "timeDelaySeconds": self.time_delay_seconds,
            "url": self.url,
            "whitelist": self.whitelist,
        }


class MediaPackageBackend(BaseBackend):
    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self._channels: Dict[str, Channel] = OrderedDict()
        self._origin_endpoints: Dict[str, OriginEndpoint] = OrderedDict()

    def create_channel(
        self, description: str, channel_id: str, tags: Dict[str, str]
    ) -> Channel:
        arn = f"arn:{get_partition(self.region_name)}:mediapackage:channel:{channel_id}"
        channel = Channel(
            arn=arn,
            description=description,
            channel_id=channel_id,
            tags=tags,
        )
        self._channels[channel_id] = channel
        return channel

    def list_channels(self) -> List[Dict[str, Any]]:
        return [c.to_dict() for c in self._channels.values()]

    def describe_channel(self, channel_id: str) -> Channel:
        try:
            return self._channels[channel_id]
        except KeyError:
            raise ClientError(
                "NotFoundException", f"channel with id={channel_id} not found"
            )

    def delete_channel(self, channel_id: str) -> Channel:
        if channel_id in self._channels:
            return self._channels.pop(channel_id)
        raise ClientError(
            "NotFoundException", f"channel with id={channel_id} not found"
        )

    def create_origin_endpoint(
        self,
        authorization: Dict[str, str],
        channel_id: str,
        cmaf_package: Dict[str, Any],
        dash_package: Dict[str, Any],
        description: str,
        hls_package: Dict[str, Any],
        endpoint_id: str,
        manifest_name: str,
        mss_package: Dict[str, Any],
        origination: str,
        startover_window_seconds: int,
        tags: Dict[str, str],
        time_delay_seconds: int,
        whitelist: List[str],
    ) -> OriginEndpoint:
        arn = f"arn:{get_partition(self.region_name)}:mediapackage:origin_endpoint:{endpoint_id}"
        url = f"https://origin-endpoint.mediapackage.{self.region_name}.amazonaws.com/{endpoint_id}"
        origin_endpoint = OriginEndpoint(
            arn=arn,
            authorization=authorization,
            channel_id=channel_id,
            cmaf_package=cmaf_package,
            dash_package=dash_package,
            description=description,
            hls_package=hls_package,
            endpoint_id=endpoint_id,
            manifest_name=manifest_name,
            mss_package=mss_package,
            origination=origination,
            startover_window_seconds=startover_window_seconds,
            tags=tags,
            time_delay_seconds=time_delay_seconds,
            url=url,
            whitelist=whitelist,
        )
        self._origin_endpoints[endpoint_id] = origin_endpoint
        return origin_endpoint

    def describe_origin_endpoint(self, endpoint_id: str) -> OriginEndpoint:
        try:
            return self._origin_endpoints[endpoint_id]
        except KeyError:
            raise ClientError(
                "NotFoundException", f"origin endpoint with id={endpoint_id} not found"
            )

    def list_origin_endpoints(self) -> List[Dict[str, Any]]:
        return [o.to_dict() for o in self._origin_endpoints.values()]

    def delete_origin_endpoint(self, endpoint_id: str) -> OriginEndpoint:
        if endpoint_id in self._origin_endpoints:
            return self._origin_endpoints.pop(endpoint_id)
        raise ClientError(
            "NotFoundException", f"origin endpoint with id={endpoint_id} not found"
        )

    def update_origin_endpoint(
        self,
        authorization: Dict[str, str],
        cmaf_package: Dict[str, Any],
        dash_package: Dict[str, Any],
        description: str,
        hls_package: Dict[str, Any],
        endpoint_id: str,
        manifest_name: str,
        mss_package: Dict[str, Any],
        origination: str,
        startover_window_seconds: int,
        time_delay_seconds: int,
        whitelist: List[str],
    ) -> OriginEndpoint:
        try:
            origin_endpoint = self._origin_endpoints[endpoint_id]
            origin_endpoint.authorization = authorization
            origin_endpoint.cmaf_package = cmaf_package
            origin_endpoint.dash_package = dash_package
            origin_endpoint.description = description
            origin_endpoint.hls_package = hls_package
            origin_endpoint.manifest_name = manifest_name
            origin_endpoint.mss_package = mss_package
            origin_endpoint.origination = origination
            origin_endpoint.startover_window_seconds = startover_window_seconds
            origin_endpoint.time_delay_seconds = time_delay_seconds
            origin_endpoint.whitelist = whitelist
            return origin_endpoint
        except KeyError:
            raise ClientError(
                "NotFoundException", f"origin endpoint with id={endpoint_id} not found"
            )


mediapackage_backends = BackendDict(MediaPackageBackend, "mediapackage")



import enum
from typing import Final

from moto.stepfunctions.parser.asl.component.component import Component


class InputTypeValue(enum.Enum):
    """
    Represents the supported InputType values for ItemReader configurations.
    """

    # TODO: add support for MANIFEST InputTypeValue.
    CSV = "CSV"
    JSON = "JSON"


class InputType(Component):
    """
    "InputType" Specifies the type of Amazon S3 data source, such as CSV file, object, JSON file, or an
    Amazon S3 inventory list. In Workflow Studio, you can select an input type from the Amazon S3 item
    source dropdown list under the Item source field.
    """

    input_type_value: Final[InputTypeValue]

    def __init__(self, input_type: str):
        self.input_type_value = InputTypeValue(input_type)  # Pass error upstream.



import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws
from moto.moto_api._internal import mock_random

from .test_route53resolver_endpoint import TEST_REGION, create_vpc
from .test_route53resolver_rule import create_test_rule


def create_test_rule_association(
    client, ec2_client, resolver_rule_id=None, name=None, vpc_id=None
):
    """Create a Resolver Rule Association for testing purposes."""
    if not resolver_rule_id:
        resolver_rule_id = create_test_rule(client)["Id"]
    name = name if name else "R" + mock_random.get_random_hex(10)
    if not vpc_id:
        vpc_id = create_vpc(ec2_client)
    return client.associate_resolver_rule(
        ResolverRuleId=resolver_rule_id, Name=name, VPCId=vpc_id
    )["ResolverRuleAssociation"]


@mock_aws
def test_route53resolver_invalid_associate_resolver_rule_args():
    """Test invalid arguments to the associate_resolver_rule API."""
    client = boto3.client("route53resolver", region_name=TEST_REGION)
    random_num = mock_random.get_random_hex(10)

    # Verify ValidationException error messages are accumulated properly:
    #  - resolver rule ID that exceeds the allowed length of 64.
    #  - name that exceeds the allowed length of 64.
    #  - vpc_id that exceeds the allowed length of 64.
    long_id = random_num * 7
    long_name = random_num * 6 + "abcde"
    long_vpc_id = random_num * 6 + "fghij"
    with pytest.raises(ClientError) as exc:
        client.associate_resolver_rule(
            ResolverRuleId=long_id, Name=long_name, VPCId=long_vpc_id
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert "3 validation errors detected" in err["Message"]
    assert (
        f"Value '{long_id}' at 'resolverRuleId' failed to satisfy "
        f"constraint: Member must have length less than or equal to 64"
    ) in err["Message"]
    assert (
        f"Value '{long_name}' at 'name' failed to satisfy constraint: "
        f"Member must have length less than or equal to 64"
    ) in err["Message"]
    assert (
        f"Value '{long_vpc_id}' at 'vPCId' failed to satisfy "
        f"constraint: Member must have length less than or equal to 64"
    ) in err["Message"]


@mock_aws
def test_route53resolver_associate_resolver_rule():
    """Test good associate_resolver_rule API calls."""
    client = boto3.client("route53resolver", region_name=TEST_REGION)
    ec2_client = boto3.client("ec2", region_name=TEST_REGION)

    resolver_rule_id = create_test_rule(client)["Id"]
    name = "X" + mock_random.get_random_hex(10)
    vpc_id = create_vpc(ec2_client)
    rule_association = client.associate_resolver_rule(
        ResolverRuleId=resolver_rule_id, Name=name, VPCId=vpc_id
    )["ResolverRuleAssociation"]
    assert rule_association["Id"].startswith("rslvr-rrassoc-")
    assert rule_association["ResolverRuleId"] == resolver_rule_id
    assert rule_association["Name"] == name
    assert rule_association["VPCId"] == vpc_id
    assert rule_association["Status"] == "COMPLETE"
    assert "StatusMessage" in rule_association


@mock_aws
def test_route53resolver_other_associate_resolver_rule_errors():
    """Test good associate_resolver_rule API calls."""
    client = boto3.client("route53resolver", region_name=TEST_REGION)
    ec2_client = boto3.client("ec2", region_name=TEST_REGION)

    # Resolver referenced by resolver_rule_id doesn't exist.
    with pytest.raises(ClientError) as exc:
        create_test_rule_association(client, ec2_client, resolver_rule_id="foo")
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"
    assert "Resolver rule with ID 'foo' does not exist" in err["Message"]

    # Invalid vpc_id
    with pytest.raises(ClientError) as exc:
        create_test_rule_association(client, ec2_client, vpc_id="foo")
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidParameterException"
    assert "The vpc ID 'foo' does not exist" in err["Message"]

    # Same resolver_rule_id and vpc_id pair for an association.
    resolver_rule_id = create_test_rule(client)["Id"]
    vpc_id = create_vpc(ec2_client)
    create_test_rule_association(
        client, ec2_client, resolver_rule_id=resolver_rule_id, vpc_id=vpc_id
    )
    with pytest.raises(ClientError) as exc:
        create_test_rule_association(
            client, ec2_client, resolver_rule_id=resolver_rule_id, vpc_id=vpc_id
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidRequestException"
    assert (
        f"Cannot associate rules with same domain name with same VPC. "
        f"Conflict with resolver rule '{resolver_rule_id}'"
    ) in err["Message"]

    # Not testing "Too many rule associations" as it takes too long to create
    # 2000 VPCs and rule associations.


@mock_aws
def test_route53resolver_disassociate_resolver_rule():
    """Test good disassociate_resolver_rule API calls."""
    client = boto3.client("route53resolver", region_name=TEST_REGION)
    ec2_client = boto3.client("ec2", region_name=TEST_REGION)
    created_association = create_test_rule_association(client, ec2_client)

    # Disassociate the resolver rule and verify the response.
    response = client.disassociate_resolver_rule(
        ResolverRuleId=created_association["ResolverRuleId"],
        VPCId=created_association["VPCId"],
    )
    association = response["ResolverRuleAssociation"]
    assert association["Id"] == created_association["Id"]
    assert association["ResolverRuleId"] == created_association["ResolverRuleId"]
    assert association["Name"] == created_association["Name"]
    assert association["VPCId"] == created_association["VPCId"]
    assert association["Status"] == "DELETING"
    assert "Deleting" in association["StatusMessage"]


@mock_aws
def test_route53resolver_bad_disassociate_resolver_rule():
    """Test disassociate_resolver_rule API calls with a bad ID."""
    client = boto3.client("route53resolver", region_name=TEST_REGION)
    ec2_client = boto3.client("ec2", region_name=TEST_REGION)
    random_num = mock_random.get_random_hex(10)

    # Use a resolver rule id and vpc id that is too long.
    long_id = "0123456789" * 6 + "xxxxx"
    long_vpc_id = random_num * 6 + "12345"
    with pytest.raises(ClientError) as exc:
        client.disassociate_resolver_rule(ResolverRuleId=long_id, VPCId=long_vpc_id)
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert "2 validation errors detected" in err["Message"]
    assert (
        f"Value '{long_id}' at 'resolverRuleId' failed to satisfy "
        f"constraint: Member must have length less than or equal to 64"
    ) in err["Message"]
    assert (
        f"Value '{long_vpc_id}' at 'vPCId' failed to satisfy "
        f"constraint: Member must have length less than or equal to 64"
    ) in err["Message"]

    # Create a test association.
    test_association = create_test_rule_association(client, ec2_client)
    test_rule_id = test_association["ResolverRuleId"]
    test_vpc_id = test_association["VPCId"]

    # Disassociate from a non-existent resolver rule id.
    with pytest.raises(ClientError) as exc:
        client.disassociate_resolver_rule(ResolverRuleId=random_num, VPCId=test_vpc_id)
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"
    assert f"Resolver rule with ID '{random_num}' does not exist" in err["Message"]

    # Disassociate using a non-existent vpc id.
    with pytest.raises(ClientError) as exc:
        client.disassociate_resolver_rule(ResolverRuleId=test_rule_id, VPCId=random_num)
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"
    assert (
        f"Resolver Rule Association between Resolver Rule "
        f"'{test_rule_id}' and VPC '{random_num}' does not exist"
    ) in err["Message"]

    # Disassociate successfully, then test that it's not possible to
    # disassociate again.
    client.disassociate_resolver_rule(ResolverRuleId=test_rule_id, VPCId=test_vpc_id)
    with pytest.raises(ClientError) as exc:
        client.disassociate_resolver_rule(
            ResolverRuleId=test_rule_id, VPCId=test_vpc_id
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"
    assert (
        f"Resolver Rule Association between Resolver Rule "
        f"'{test_rule_id}' and VPC '{test_vpc_id}' does not exist"
    ) in err["Message"]


@mock_aws
def test_route53resolver_get_resolver_rule_association():
    """Test good get_resolver_rule_association API calls."""
    client = boto3.client("route53resolver", region_name=TEST_REGION)
    ec2_client = boto3.client("ec2", region_name=TEST_REGION)

    # Create a good association for testing purposes.
    created_association = create_test_rule_association(client, ec2_client)

    # Now get the resolver rule association and verify the response.
    response = client.get_resolver_rule_association(
        ResolverRuleAssociationId=created_association["Id"]
    )
    association = response["ResolverRuleAssociation"]
    assert association["Id"] == created_association["Id"]
    assert association["ResolverRuleId"] == created_association["ResolverRuleId"]
    assert association["Name"] == created_association["Name"]
    assert association["VPCId"] == created_association["VPCId"]
    assert association["Status"] == created_association["Status"]
    assert association["StatusMessage"] == created_association["StatusMessage"]


@mock_aws
def test_route53resolver_bad_get_resolver_rule_association():
    """Test get_resolver_rule_association API calls with a bad ID."""
    client = boto3.client("route53resolver", region_name=TEST_REGION)
    random_num = mock_random.get_random_hex(10)

    # Use a resolver rule association id that is too long.
    long_id = "0123456789" * 6 + "xxxxx"
    with pytest.raises(ClientError) as exc:
        client.get_resolver_rule_association(ResolverRuleAssociationId=long_id)
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert "1 validation error detected" in err["Message"]
    assert (
        f"Value '{long_id}' at 'resolverRuleAssociationId' failed to satisfy "
        f"constraint: Member must have length less than or equal to 64"
    ) in err["Message"]

    # Get a non-existent resolver rule association.
    with pytest.raises(ClientError) as exc:
        client.get_resolver_rule_association(ResolverRuleAssociationId=random_num)
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"
    assert f"ResolverRuleAssociation '{random_num}' does not Exist" in err["Message"]


@mock_aws
def test_route53resolver_list_resolver_rule_associations():
    """Test good list_resolver_rule_associations API calls."""
    client = boto3.client("route53resolver", region_name=TEST_REGION)
    ec2_client = boto3.client("ec2", region_name=TEST_REGION)
    random_num = mock_random.get_random_hex(10)

    # List rule associations when there are none.
    response = client.list_resolver_rule_associations()
    assert len(response["ResolverRuleAssociations"]) == 0
    assert response["MaxResults"] == 10
    assert "NextToken" not in response

    # Create 4 associations, verify all are listed when no filters, max_results.
    for idx in range(4):
        create_test_rule_association(client, ec2_client, name=f"A{idx}-{random_num}")
    response = client.list_resolver_rule_associations()
    associations = response["ResolverRuleAssociations"]
    assert len(associations) == 4
    assert response["MaxResults"] == 10
    for idx in range(4):
        assert associations[idx]["Name"].startswith(f"A{idx}")

    # Set max_results to return 1 association, use next_token to get
    # remaining 3.
    response = client.list_resolver_rule_associations(MaxResults=1)
    associations = response["ResolverRuleAssociations"]
    assert len(associations) == 1
    assert response["MaxResults"] == 1
    assert "NextToken" in response
    assert associations[0]["Name"].startswith("A0")

    response = client.list_resolver_rule_associations(NextToken=response["NextToken"])
    associations = response["ResolverRuleAssociations"]
    assert len(associations) == 3
    assert response["MaxResults"] == 10
    assert "NextToken" not in response
    for idx, association in enumerate(associations):
        assert association["Name"].startswith(f"A{idx + 1}")


@mock_aws
def test_route53resolver_list_resolver_rule_associations_filters():
    """Test good list_resolver_rule_associations API calls that use filters."""
    client = boto3.client("route53resolver", region_name=TEST_REGION)
    ec2_client = boto3.client("ec2", region_name=TEST_REGION)
    random_num = mock_random.get_random_hex(10)

    # Create some rule associations for testing purposes
    vpc_id1 = create_vpc(ec2_client)
    vpc_id2 = create_vpc(ec2_client)
    associations = []
    rule_ids = ["zero_offset"]
    for idx in range(1, 5):
        association = create_test_rule_association(
            client,
            ec2_client,
            name=f"F{idx}-{random_num}",
            vpc_id=vpc_id1 if idx % 2 else vpc_id2,
        )
        associations.append(association)
        rule_ids.append(association["ResolverRuleId"])

    # Try all the valid filter names, including some of the old style names.
    response = client.list_resolver_rule_associations(
        Filters=[{"Name": "ResolverRuleId", "Values": [rule_ids[3]]}]
    )
    assert len(response["ResolverRuleAssociations"]) == 1
    assert response["ResolverRuleAssociations"][0]["ResolverRuleId"] == rule_ids[3]

    response = client.list_resolver_rule_associations(
        Filters=[{"Name": "RESOLVER_RULE_ID", "Values": [rule_ids[2], rule_ids[4]]}]
    )
    assert len(response["ResolverRuleAssociations"]) == 2
    assert response["ResolverRuleAssociations"][0]["ResolverRuleId"] == rule_ids[2]
    assert response["ResolverRuleAssociations"][1]["ResolverRuleId"] == rule_ids[4]

    response = client.list_resolver_rule_associations(
        Filters=[{"Name": "VPCId", "Values": [vpc_id2]}]
    )
    assert len(response["ResolverRuleAssociations"]) == 2

    response = client.list_resolver_rule_associations(
        Filters=[{"Name": "Name", "Values": [f"F1-{random_num}"]}]
    )
    assert len(response["ResolverRuleAssociations"]) == 1
    assert response["ResolverRuleAssociations"][0]["Name"] == f"F1-{random_num}"

    response = client.list_resolver_rule_associations(
        Filters=[
            {"Name": "VPC_ID", "Values": [vpc_id1]},
            {"Name": "NAME", "Values": [f"F3-{random_num}"]},
        ]
    )
    assert len(response["ResolverRuleAssociations"]) == 1
    assert response["ResolverRuleAssociations"][0]["Name"] == f"F3-{random_num}"

    response = client.list_resolver_rule_associations(
        Filters=[{"Name": "Status", "Values": ["COMPLETE"]}]
    )
    assert len(response["ResolverRuleAssociations"]) == 4
    response = client.list_resolver_rule_associations(
        Filters=[{"Name": "Status", "Values": ["CREATING"]}]
    )
    assert len(response["ResolverRuleAssociations"]) == 0


@mock_aws
def test_route53resolver_bad_list_resolver_rule_associations_filters():
    """Test bad list_resolver_rule_associations API calls that use filters."""
    client = boto3.client("route53resolver", region_name=TEST_REGION)

    # botocore barfs on an empty "Values":
    # TypeError: list_resolver_rule_associations() only accepts keyword arguments.
    # client.list_resolver_rule_associations([{"Name": "Direction", "Values": []}])
    # client.list_resolver_rule_associations([{"Values": []}])

    with pytest.raises(ClientError) as exc:
        client.list_resolver_rule_associations(
            Filters=[{"Name": "foo", "Values": ["bar"]}]
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidParameterException"
    assert "The filter 'foo' is invalid" in err["Message"]

    with pytest.raises(ClientError) as exc:
        client.list_resolver_rule_associations(
            Filters=[{"Name": "VpcId", "Values": ["bar"]}]
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidParameterException"
    assert "The filter 'VpcId' is invalid" in err["Message"]


@mock_aws
def test_route53resolver_bad_list_resolver_rule_associations():
    """Test bad list_resolver_rule_associations API calls."""
    client = boto3.client("route53resolver", region_name=TEST_REGION)
    ec2_client = boto3.client("ec2", region_name=TEST_REGION)

    # Bad max_results.
    create_test_rule_association(client, ec2_client)
    with pytest.raises(ClientError) as exc:
        client.list_resolver_rule_associations(MaxResults=250)
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert "1 validation error detected" in err["Message"]
    assert (
        "Value '250' at 'maxResults' failed to satisfy constraint: Member "
        "must have length less than or equal to 100"
    ) in err["Message"]



"""rdsdata base URL and path."""

from .responses import RDSDataServiceResponse

url_bases = [
    r"https?://rds-data\.(.+)\.amazonaws\.com",
]


response = RDSDataServiceResponse()


url_paths = {
    "{0}/Execute$": response.dispatch,
}



from .responses import GuardDutyResponse

url_bases = [
    "https?://guardduty\\.(.+)\\.amazonaws\\.com",
]


url_paths = {
    "{0}/detector$": GuardDutyResponse.dispatch,
    "{0}/detector/(?P<detector_id>[^/]+)$": GuardDutyResponse.dispatch,
    "{0}/detector/(?P<detector_id>[^/]+)/filter$": GuardDutyResponse.dispatch,
    "{0}/detector/(?P<detector_id>[^/]+)/filter/(?P<filter_name>[^/]+)$": GuardDutyResponse.dispatch,
    "{0}/admin/enable$": GuardDutyResponse.dispatch,
    "{0}/admin$": GuardDutyResponse.dispatch,
}



import json
from functools import wraps
from time import sleep
from uuid import uuid4

import boto3
import pytest
import requests

from moto import mock_aws
from tests.test_awslambda.utilities import get_test_zip_file_error

from ...markers import requires_docker
from . import base_url, sfn_role_policy
from .templates.retry import retry_template
from .templates.retry_jitter_none import retry_template_jitter_none


def create_failing_lambda_func(func):
    @wraps(func)
    def pagination_wrapper():
        role_name = "moto_test_role_" + str(uuid4())[0:6]

        with mock_aws():
            requests.post(
                f"http://{base_url}/moto-api/config",
                json={"stepfunctions": {"execute_state_machine": True}},
            )
            resp = create_role_and_test(role_name)
            requests.post(
                f"http://{base_url}/moto-api/config",
                json={"stepfunctions": {"execute_state_machine": False}},
            )
            return resp

    def create_role_and_test(role_name):
        policy_doc = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "LambdaAssumeRole",
                    "Effect": "Allow",
                    "Principal": {"Service": "lambda.amazonaws.com"},
                    "Action": "sts:AssumeRole",
                }
            ],
        }
        iam = boto3.client("iam", region_name="us-east-1")
        iam_role_arn = iam.create_role(
            RoleName=role_name,
            AssumeRolePolicyDocument=json.dumps(policy_doc),
            Path="/",
        )["Role"]["Arn"]

        fn_arn = create_function(iam_role_arn)

        return func(fn_arn)

    def create_function(role_arn: str):
        _lambda = boto3.client("lambda", "us-east-1")
        fn = _lambda.create_function(
            FunctionName=f"fn_{str(uuid4())[0:6]}",
            Runtime="python3.11",
            Role=role_arn,
            Handler="lambda_function.lambda_handler",
            Code={"ZipFile": get_test_zip_file_error()},
        )
        return fn["FunctionArn"]

    return pagination_wrapper


@create_failing_lambda_func
@pytest.mark.network
@requires_docker
def test_retry_lambda(fn_arn=None):
    definition = retry_template.copy()
    definition["States"]["LambdaTask"]["Resource"] = fn_arn

    iam = boto3.client("iam", region_name="us-east-1")
    role_name = f"sfn_role_{str(uuid4())[0:6]}"
    sfn_role = iam.create_role(
        RoleName=role_name,
        AssumeRolePolicyDocument=json.dumps(sfn_role_policy),
        Path="/",
    )["Role"]["Arn"]

    client = boto3.client("stepfunctions", region_name="us-east-1")
    name = "sfn_" + str(uuid4())[0:6]
    exec_input = {"my": "input"}
    #
    response = client.create_state_machine(
        name=name, definition=json.dumps(definition), roleArn=sfn_role
    )
    state_machine_arn = response["stateMachineArn"]

    execution = client.start_execution(
        name="exec1", stateMachineArn=state_machine_arn, input=json.dumps(exec_input)
    )
    execution_arn = execution["executionArn"]

    for _ in range(25):
        execution = client.describe_execution(executionArn=execution_arn)
        if execution["status"] == "FAILED":
            history = client.get_execution_history(executionArn=execution_arn)
            assert len(history["events"]) == 12
            assert [e["type"] for e in history["events"]] == [
                "ExecutionStarted",
                "TaskStateEntered",
                "LambdaFunctionScheduled",
                "LambdaFunctionStarted",
                "LambdaFunctionFailed",
                "LambdaFunctionScheduled",
                "LambdaFunctionStarted",
                "LambdaFunctionFailed",
                "LambdaFunctionScheduled",
                "LambdaFunctionStarted",
                "LambdaFunctionFailed",
                "ExecutionFailed",
            ]

            break
        sleep(1)
    else:
        assert False, "Should have failed already"


@create_failing_lambda_func
@pytest.mark.network
@requires_docker
def test_retry_lambda_jitter_none(fn_arn=None):
    definition = retry_template_jitter_none.copy()
    definition["States"]["LambdaTask"]["Resource"] = fn_arn

    iam = boto3.client("iam", region_name="us-east-1")
    role_name = f"sfn_role_{str(uuid4())[0:6]}"
    sfn_role = iam.create_role(
        RoleName=role_name,
        AssumeRolePolicyDocument=json.dumps(sfn_role_policy),
        Path="/",
    )["Role"]["Arn"]

    client = boto3.client("stepfunctions", region_name="us-east-1")
    name = "sfn_" + str(uuid4())[0:6]
    exec_input = {"my": "input"}
    #
    response = client.create_state_machine(
        name=name, definition=json.dumps(definition), roleArn=sfn_role
    )
    state_machine_arn = response["stateMachineArn"]

    execution = client.start_execution(
        name="exec1", stateMachineArn=state_machine_arn, input=json.dumps(exec_input)
    )
    execution_arn = execution["executionArn"]

    for _ in range(25):
        execution = client.describe_execution(executionArn=execution_arn)
        if execution["status"] == "FAILED":
            history = client.get_execution_history(executionArn=execution_arn)
            assert len(history["events"]) == 12
            assert [e["type"] for e in history["events"]] == [
                "ExecutionStarted",
                "TaskStateEntered",
                "LambdaFunctionScheduled",
                "LambdaFunctionStarted",
                "LambdaFunctionFailed",
                "LambdaFunctionScheduled",
                "LambdaFunctionStarted",
                "LambdaFunctionFailed",
                "LambdaFunctionScheduled",
                "LambdaFunctionStarted",
                "LambdaFunctionFailed",
                "ExecutionFailed",
            ]

            break
        sleep(1)
    else:
        assert False, "Should have failed already"



import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws
from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID


def create_stream(client):
    stream_name = "my-stream"
    client.create_stream(StreamName=stream_name, ShardCount=4)
    stream = client.describe_stream(StreamName=stream_name)["StreamDescription"]
    return stream["StreamARN"]


@mock_aws
def test_list_stream_consumers():
    client = boto3.client("kinesis", region_name="eu-west-1")
    stream_arn = create_stream(client)

    resp = client.list_stream_consumers(StreamARN=stream_arn)

    assert resp["Consumers"] == []


@mock_aws
def test_register_stream_consumer():
    client = boto3.client("kinesis", region_name="eu-west-1")
    stream_arn = create_stream(client)

    resp = client.register_stream_consumer(
        StreamARN=stream_arn, ConsumerName="newconsumer"
    )
    assert "Consumer" in resp

    consumer = resp["Consumer"]

    assert consumer["ConsumerName"] == "newconsumer"
    assert (
        consumer["ConsumerARN"]
        == f"arn:aws:kinesis:eu-west-1:{ACCOUNT_ID}:stream/my-stream/consumer/newconsumer"
    )
    assert consumer["ConsumerStatus"] == "ACTIVE"
    assert "ConsumerCreationTimestamp" in consumer

    resp = client.list_stream_consumers(StreamARN=stream_arn)

    assert len(resp["Consumers"]) == 1
    consumer = resp["Consumers"][0]
    assert consumer["ConsumerName"] == "newconsumer"
    assert (
        consumer["ConsumerARN"]
        == f"arn:aws:kinesis:eu-west-1:{ACCOUNT_ID}:stream/my-stream/consumer/newconsumer"
    )
    assert consumer["ConsumerStatus"] == "ACTIVE"
    assert "ConsumerCreationTimestamp" in consumer


@mock_aws
def test_describe_stream_consumer_by_name():
    client = boto3.client("kinesis", region_name="us-east-2")
    stream_arn = create_stream(client)
    client.register_stream_consumer(StreamARN=stream_arn, ConsumerName="newconsumer")

    resp = client.describe_stream_consumer(
        StreamARN=stream_arn, ConsumerName="newconsumer"
    )
    assert "ConsumerDescription" in resp

    consumer = resp["ConsumerDescription"]
    assert consumer["ConsumerName"] == "newconsumer"
    assert "ConsumerARN" in consumer
    assert consumer["ConsumerStatus"] == "ACTIVE"
    assert "ConsumerCreationTimestamp" in consumer
    assert consumer["StreamARN"] == stream_arn


@mock_aws
def test_describe_stream_consumer_by_arn():
    client = boto3.client("kinesis", region_name="us-east-2")
    stream_arn = create_stream(client)
    resp = client.register_stream_consumer(
        StreamARN=stream_arn, ConsumerName="newconsumer"
    )
    consumer_arn = resp["Consumer"]["ConsumerARN"]

    resp = client.describe_stream_consumer(ConsumerARN=consumer_arn)
    assert "ConsumerDescription" in resp

    consumer = resp["ConsumerDescription"]
    assert consumer["ConsumerName"] == "newconsumer"
    assert "ConsumerARN" in consumer
    assert consumer["ConsumerStatus"] == "ACTIVE"
    assert "ConsumerCreationTimestamp" in consumer
    assert consumer["StreamARN"] == stream_arn


@mock_aws
def test_describe_stream_consumer_unknown():
    client = boto3.client("kinesis", region_name="us-east-2")
    create_stream(client)

    unknown_arn = f"arn:aws:kinesis:us-east-2:{ACCOUNT_ID}:stream/unknown"
    with pytest.raises(ClientError) as exc:
        client.describe_stream_consumer(ConsumerARN=unknown_arn)
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"
    assert err["Message"] == f"Consumer {unknown_arn}, account {ACCOUNT_ID} not found."


@mock_aws
def test_deregister_stream_consumer_by_name():
    client = boto3.client("kinesis", region_name="ap-southeast-1")
    stream_arn = create_stream(client)

    client.register_stream_consumer(StreamARN=stream_arn, ConsumerName="consumer1")
    client.register_stream_consumer(StreamARN=stream_arn, ConsumerName="consumer2")

    assert len(client.list_stream_consumers(StreamARN=stream_arn)["Consumers"]) == 2

    client.deregister_stream_consumer(StreamARN=stream_arn, ConsumerName="consumer1")

    assert len(client.list_stream_consumers(StreamARN=stream_arn)["Consumers"]) == 1


@mock_aws
def test_deregister_stream_consumer_by_arn():
    client = boto3.client("kinesis", region_name="ap-southeast-1")
    stream_arn = create_stream(client)

    resp = client.register_stream_consumer(
        StreamARN=stream_arn, ConsumerName="consumer1"
    )
    consumer1_arn = resp["Consumer"]["ConsumerARN"]
    client.register_stream_consumer(StreamARN=stream_arn, ConsumerName="consumer2")

    assert len(client.list_stream_consumers(StreamARN=stream_arn)["Consumers"]) == 2

    client.deregister_stream_consumer(ConsumerARN=consumer1_arn)

    assert len(client.list_stream_consumers(StreamARN=stream_arn)["Consumers"]) == 1



from moto.utilities.utils import get_partition


def make_arn_for_dashboard(account_id: str, region_name: str, name: str) -> str:
    return f"arn:{get_partition(region_name)}:cloudwatch::{account_id}:dashboard/{name}"


def make_arn_for_alarm(region: str, account_id: str, alarm_name: str) -> str:
    return f"arn:{get_partition(region)}:cloudwatch:{region}:{account_id}:alarm:{alarm_name}"



from typing import Any, Dict, Optional

from moto.core.responses import BaseResponse

from ..exceptions import EmptyTagSpecError, InvalidParameter
from ..utils import convert_tag_spec


class EC2BaseResponse(BaseResponse):
    @property
    def ec2_backend(self) -> Any:  # type: ignore[misc]
        from moto.ec2.models import ec2_backends

        return ec2_backends[self.current_account][self.region]

    def _filters_from_querystring(self) -> Dict[str, str]:
        # [{"Name": x1, "Value": y1}, ..]
        _filters = self._get_multi_param("Filter.")
        # return {x1: y1, ...}
        return {f["Name"]: f["Value"] for f in _filters}

    def _parse_tag_specification(
        self, expected_type: Optional[str] = None
    ) -> Dict[str, Dict[str, str]]:
        # [{"ResourceType": _type, "Tag": [{"Key": k, "Value": v}, ..]}]
        tag_spec_set = self._get_multi_param(
            "TagSpecification", skip_result_conversion=True
        )
        if not tag_spec_set:
            tag_spec_set = self._get_multi_param(
                "TagSpecifications", skip_result_conversion=True
            )
        if not tag_spec_set:
            return {}

        tags_dict = (
            tag_spec_set[0] if isinstance(tag_spec_set, list) else tag_spec_set
        )  # awscli allows for a json string to be passed and it should be allowed
        if "ResourceType" not in tags_dict:
            raise InvalidParameter("Tag specification resource type must have a value")
        if expected_type and tags_dict["ResourceType"] != expected_type:
            raise InvalidParameter(
                f"'{tags_dict['ResourceType']}' is not a valid taggable resource type for this operation."
            )
        if "Tag" not in tags_dict:
            if tags_dict.get("ResourceType") == "subnet":
                raise InvalidParameter("Tag specification must have at least one tag")
            raise EmptyTagSpecError

        return convert_tag_spec(tag_spec_set)



from typing import Any

from moto.core.common_models import BaseModel
from moto.core.utils import unix_time


class Timeout(BaseModel):
    def __init__(self, obj: Any, timestamp: float, kind: str):
        self.obj = obj
        self.timestamp = timestamp
        self.kind = kind

    @property
    def reached(self) -> bool:
        return unix_time() >= self.timestamp



import re
from collections import defaultdict
from typing import Dict, List, Optional

from ..exceptions import (
    InvalidParameterValueErrorTagNull,
    TagLimitExceeded,
)
from ..utils import (
    EC2_PREFIX_TO_RESOURCE,
    get_prefix,
    simple_aws_filter_to_re,
)


class TagBackend:
    VALID_TAG_FILTERS = ["key", "resource-id", "resource-type", "value"]

    def __init__(self) -> None:
        self.tags: Dict[str, Dict[str, str]] = defaultdict(dict)

    def create_tags(self, resource_ids: List[str], tags: Dict[str, str]) -> bool:
        if None in set([tags[tag] for tag in tags]):
            raise InvalidParameterValueErrorTagNull()
        for resource_id in resource_ids:
            if resource_id in self.tags:
                if (
                    len(self.tags[resource_id])
                    + len([tag for tag in tags if not tag.startswith("aws:")])
                    > 50
                ):
                    raise TagLimitExceeded()
            elif len([tag for tag in tags if not tag.startswith("aws:")]) > 50:
                raise TagLimitExceeded()
        for resource_id in resource_ids:
            for tag in tags:
                self.tags[resource_id][tag] = tags[tag]
        return True

    def delete_tags(self, resource_ids: List[str], tags: Dict[str, str]) -> bool:
        for resource_id in resource_ids:
            for tag in tags:
                if tag in self.tags[resource_id]:
                    if tags[tag] is None:
                        self.tags[resource_id].pop(tag)
                    elif tags[tag] == self.tags[resource_id][tag]:
                        self.tags[resource_id].pop(tag)
        return True

    def describe_tags(
        self, filters: Optional[Dict[str, List[str]]] = None
    ) -> List[Dict[str, str]]:
        results = []
        key_filters = []
        resource_id_filters = []
        resource_type_filters = []
        value_filters = []
        if filters is not None:
            for tag_filter in filters:
                if tag_filter in self.VALID_TAG_FILTERS:
                    if tag_filter == "key":
                        for value in filters[tag_filter]:
                            key_filters.append(
                                re.compile(simple_aws_filter_to_re(value))
                            )
                    if tag_filter == "resource-id":
                        for value in filters[tag_filter]:
                            resource_id_filters.append(
                                re.compile(simple_aws_filter_to_re(value))
                            )
                    if tag_filter == "resource-type":
                        for value in filters[tag_filter]:
                            resource_type_filters.append(value)
                    if tag_filter == "value":
                        for value in filters[tag_filter]:
                            value_filters.append(
                                re.compile(simple_aws_filter_to_re(value))
                            )
        for resource_id, tags in self.tags.copy().items():
            for key, value in tags.items():
                add_result = False
                if filters is None:
                    add_result = True
                else:
                    key_pass = False
                    id_pass = False
                    type_pass = False
                    value_pass = False
                    if key_filters:
                        for pattern in key_filters:
                            if pattern.match(key) is not None:
                                key_pass = True
                    else:
                        key_pass = True
                    if resource_id_filters:
                        for pattern in resource_id_filters:
                            if pattern.match(resource_id) is not None:
                                id_pass = True
                    else:
                        id_pass = True
                    if resource_type_filters:
                        for resource_type in resource_type_filters:
                            if (
                                EC2_PREFIX_TO_RESOURCE[get_prefix(resource_id)]
                                == resource_type
                            ):
                                type_pass = True
                    else:
                        type_pass = True
                    if value_filters:
                        for pattern in value_filters:
                            if pattern.match(value) is not None:
                                value_pass = True
                    else:
                        value_pass = True
                    if key_pass and id_pass and type_pass and value_pass:
                        add_result = True
                        # If we're not filtering, or we are filtering and this
                if add_result:
                    result = {
                        "resource_id": resource_id,
                        "key": key,
                        "value": value,
                        "resource_type": EC2_PREFIX_TO_RESOURCE.get(
                            get_prefix(resource_id), ""
                        ),
                    }
                    results.append(result)
        return results



import datetime
from collections import defaultdict
from typing import Any, Dict, List, Optional

from dateutil import parser

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.core.utils import iso_8601_datetime_with_milliseconds
from moto.moto_api._internal import mock_random
from moto.utilities.utils import get_partition


class CodeBuildProjectMetadata(BaseModel):
    def __init__(
        self,
        account_id: str,
        region_name: str,
        project_name: str,
        source_version: Optional[str],
        artifacts: Optional[Dict[str, Any]],
        build_id: str,
        service_role: str,
    ):
        current_date = iso_8601_datetime_with_milliseconds()
        self.build_metadata: Dict[str, Any] = dict()

        self.build_metadata["id"] = build_id
        self.build_metadata["arn"] = (
            f"arn:{get_partition(region_name)}:codebuild:{region_name}:{account_id}:build/{build_id}"
        )

        self.build_metadata["buildNumber"] = mock_random.randint(1, 100)
        self.build_metadata["startTime"] = current_date
        self.build_metadata["currentPhase"] = "QUEUED"
        self.build_metadata["buildStatus"] = "IN_PROGRESS"
        self.build_metadata["sourceVersion"] = (
            source_version if source_version else "refs/heads/main"
        )
        self.build_metadata["projectName"] = project_name

        self.build_metadata["phases"] = [
            {
                "phaseType": "SUBMITTED",
                "phaseStatus": "SUCCEEDED",
                "startTime": current_date,
                "endTime": current_date,
                "durationInSeconds": 0,
            },
            {"phaseType": "QUEUED", "startTime": current_date},
        ]

        self.build_metadata["source"] = {
            "type": "CODECOMMIT",  # should be different based on what you pass in
            "location": "https://git-codecommit.eu-west-2.amazonaws.com/v1/repos/testing",
            "gitCloneDepth": 1,
            "gitSubmodulesConfig": {"fetchSubmodules": False},
            "buildspec": "buildspec/stuff.yaml",  # should present in the codebuild project somewhere
            "insecureSsl": False,
        }

        self.build_metadata["secondarySources"] = []
        self.build_metadata["secondarySourceVersions"] = []
        self.build_metadata["artifacts"] = artifacts
        self.build_metadata["secondaryArtifacts"] = []
        self.build_metadata["cache"] = {"type": "NO_CACHE"}

        self.build_metadata["environment"] = {
            "type": "LINUX_CONTAINER",
            "image": "aws/codebuild/amazonlinux2-x86_64-standard:3.0",
            "computeType": "BUILD_GENERAL1_SMALL",
            "environmentVariables": [],
            "privilegedMode": False,
            "imagePullCredentialsType": "CODEBUILD",
        }

        self.build_metadata["serviceRole"] = service_role

        self.build_metadata["logs"] = {
            "deepLink": "https://console.aws.amazon.com/cloudwatch/home?region=eu-west-2#logEvent:group=null;stream=null",
            "cloudWatchLogsArn": f"arn:{get_partition(region_name)}:logs:{region_name}:{account_id}:log-group:null:log-stream:null",
            "cloudWatchLogs": {"status": "ENABLED"},
            "s3Logs": {"status": "DISABLED", "encryptionDisabled": False},
        }

        self.build_metadata["timeoutInMinutes"] = 45
        self.build_metadata["queuedTimeoutInMinutes"] = 480
        self.build_metadata["buildComplete"] = False
        self.build_metadata["initiator"] = "rootme"
        self.build_metadata["encryptionKey"] = (
            f"arn:{get_partition(region_name)}:kms:{region_name}:{account_id}:alias/aws/s3"
        )


class CodeBuild(BaseModel):
    def __init__(
        self,
        account_id: str,
        region: str,
        project_name: str,
        project_source: Dict[str, Any],
        artifacts: Dict[str, Any],
        environment: Dict[str, Any],
        serviceRole: str = "some_role",
    ):
        current_date = iso_8601_datetime_with_milliseconds()
        self.project_metadata: Dict[str, Any] = dict()

        self.project_metadata["name"] = project_name
        self.project_metadata["arn"] = (
            f"arn:{get_partition(region)}:codebuild:{region}:{account_id}:project/{project_name}"
        )
        self.project_metadata["encryptionKey"] = (
            f"arn:{get_partition(region)}:kms:{region}:{account_id}:alias/aws/s3"
        )
        self.project_metadata["serviceRole"] = (
            f"arn:{get_partition(region)}:iam::{account_id}:role/service-role/{serviceRole}"
        )
        self.project_metadata["lastModifiedDate"] = current_date
        self.project_metadata["created"] = current_date
        self.project_metadata["badge"] = dict()
        self.project_metadata["badge"]["badgeEnabled"] = (
            False  # this false needs to be a json false not a python false
        )
        self.project_metadata["environment"] = environment
        self.project_metadata["artifacts"] = artifacts
        self.project_metadata["source"] = project_source
        self.project_metadata["cache"] = dict()
        self.project_metadata["cache"]["type"] = "NO_CACHE"
        self.project_metadata["timeoutInMinutes"] = ""
        self.project_metadata["queuedTimeoutInMinutes"] = ""


class CodeBuildBackend(BaseBackend):
    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.codebuild_projects: Dict[str, CodeBuild] = dict()
        self.build_history: Dict[str, List[str]] = dict()
        self.build_metadata: Dict[str, CodeBuildProjectMetadata] = dict()
        self.build_metadata_history: Dict[str, List[Dict[str, Any]]] = defaultdict(list)

    def create_project(
        self,
        project_name: str,
        project_source: Dict[str, Any],
        artifacts: Dict[str, Any],
        environment: Dict[str, Any],
        service_role: str,
    ) -> Dict[str, Any]:
        # required in other functions that don't
        self.project_name = project_name
        self.service_role = service_role

        self.codebuild_projects[project_name] = CodeBuild(
            self.account_id,
            self.region_name,
            project_name,
            project_source,
            artifacts,
            environment,
            service_role,
        )

        # empty build history
        self.build_history[project_name] = list()

        return self.codebuild_projects[project_name].project_metadata

    def list_projects(self) -> List[str]:
        projects = []

        for project in self.codebuild_projects.keys():
            projects.append(project)

        return projects

    def start_build(
        self,
        project_name: str,
        source_version: Optional[str] = None,
        artifact_override: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        build_id = f"{project_name}:{mock_random.uuid4()}"

        # construct a new build
        self.build_metadata[project_name] = CodeBuildProjectMetadata(
            self.account_id,
            self.region_name,
            project_name,
            source_version,
            artifact_override,
            build_id,
            self.service_role,
        )

        self.build_history[project_name].append(build_id)

        # update build histroy with metadata for build id
        self.build_metadata_history[project_name].append(
            self.build_metadata[project_name].build_metadata
        )

        return self.build_metadata[project_name].build_metadata

    def _set_phases(self, phases: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        current_date = iso_8601_datetime_with_milliseconds()
        # No phaseStatus for QUEUED on first start
        for existing_phase in phases:
            if existing_phase["phaseType"] == "QUEUED":
                existing_phase["phaseStatus"] = "SUCCEEDED"

        statuses = [
            "PROVISIONING",
            "DOWNLOAD_SOURCE",
            "INSTALL",
            "PRE_BUILD",
            "BUILD",
            "POST_BUILD",
            "UPLOAD_ARTIFACTS",
            "FINALIZING",
            "COMPLETED",
        ]

        for status in statuses:
            phase: Dict[str, Any] = dict()
            phase["phaseType"] = status
            phase["phaseStatus"] = "SUCCEEDED"
            phase["startTime"] = current_date
            phase["endTime"] = current_date
            phase["durationInSeconds"] = mock_random.randint(10, 100)
            phases.append(phase)

        return phases

    def batch_get_builds(self, ids: List[str]) -> List[Dict[str, Any]]:
        batch_build_metadata: List[Dict[str, Any]] = []

        for metadata in self.build_metadata_history.values():
            for build in metadata:
                if build["id"] in ids:
                    build["phases"] = self._set_phases(build["phases"])
                    build["endTime"] = iso_8601_datetime_with_milliseconds(
                        parser.parse(build["startTime"])
                        + datetime.timedelta(minutes=mock_random.randint(1, 5))
                    )
                    build["currentPhase"] = "COMPLETED"
                    build["buildStatus"] = "SUCCEEDED"

                    batch_build_metadata.append(build)

        return batch_build_metadata

    def list_builds_for_project(self, project_name: str) -> List[str]:
        try:
            return self.build_history[project_name]
        except KeyError:
            return list()

    def list_builds(self) -> List[str]:
        ids = []

        for build_ids in self.build_history.values():
            ids += build_ids
        return ids

    def delete_project(self, project_name: str) -> None:
        self.build_metadata.pop(project_name, None)
        self.codebuild_projects.pop(project_name, None)

    def stop_build(self, build_id: str) -> Optional[Dict[str, Any]]:  # type: ignore[return]
        for metadata in self.build_metadata_history.values():
            for build in metadata:
                if build["id"] == build_id:
                    # set completion properties with variable completion time
                    build["phases"] = self._set_phases(build["phases"])
                    build["endTime"] = iso_8601_datetime_with_milliseconds(
                        parser.parse(build["startTime"])
                        + datetime.timedelta(minutes=mock_random.randint(1, 5))
                    )
                    build["currentPhase"] = "COMPLETED"
                    build["buildStatus"] = "STOPPED"

                    return build


codebuild_backends = BackendDict(CodeBuildBackend, "codebuild")



from __future__ import annotations

from typing import Final, Optional

from moto.stepfunctions.parser.asl.component.common.comment import Comment
from moto.stepfunctions.parser.asl.component.common.flow.start_at import StartAt
from moto.stepfunctions.parser.asl.component.state.exec.state_map.item_reader.item_reader_decl import (
    ItemReader,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.item_selector import (
    ItemSelector,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.distributed_iteration_component import (
    DistributedIterationComponent,
    DistributedIterationComponentEvalInput,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.itemprocessor.distributed_item_processor_worker import (
    DistributedItemProcessorWorker,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.itemprocessor.processor_config import (
    ProcessorConfig,
)
from moto.stepfunctions.parser.asl.component.states import States
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.parse.typed_props import TypedProps


class DistributedItemProcessorEvalInput(DistributedIterationComponentEvalInput):
    item_selector: Final[Optional[ItemSelector]]

    def __init__(
        self,
        state_name: str,
        max_concurrency: int,
        item_reader: ItemReader,
        item_selector: Optional[ItemSelector],
    ):
        super().__init__(
            state_name=state_name,
            max_concurrency=max_concurrency,
            item_reader=item_reader,
        )
        self.item_selector = item_selector


class DistributedItemProcessor(DistributedIterationComponent):
    _processor_config: Final[ProcessorConfig]
    _eval_input: Optional[DistributedItemProcessorEvalInput]

    def __init__(
        self,
        start_at: StartAt,
        states: States,
        comment: Optional[Comment],
        processor_config: ProcessorConfig,
    ):
        super().__init__(start_at=start_at, states=states, comment=comment)
        self._processor_config = processor_config

    @classmethod
    def from_props(cls, props: TypedProps) -> DistributedItemProcessor:
        item_processor = cls(
            start_at=props.get(
                typ=StartAt,
                raise_on_missing=ValueError(
                    f"Missing StartAt declaration in props '{props}'."
                ),
            ),
            states=props.get(
                typ=States,
                raise_on_missing=ValueError(
                    f"Missing States declaration in props '{props}'."
                ),
            ),
            comment=props.get(Comment),
            processor_config=props.get(ProcessorConfig),
        )
        return item_processor

    def _create_worker(self, env: Environment) -> DistributedItemProcessorWorker:
        return DistributedItemProcessorWorker(
            work_name=self._eval_input.state_name,
            job_pool=self._job_pool,
            env=env,
            item_reader=self._eval_input.item_reader,
            item_selector=self._eval_input.item_selector,
            map_run_record=self._map_run_record,
        )



import json
from contextlib import contextmanager
from datetime import datetime
from time import sleep
from unittest import SkipTest

import boto3
import botocore
import pytest

from moto import mock_aws, settings
from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID
from moto.sagemaker.exceptions import ValidationError
from moto.sagemaker.models import FakePipeline, sagemaker_backends
from moto.sagemaker.utils import (
    arn_formatter,
    get_pipeline_execution_from_arn,
    get_pipeline_from_name,
    get_pipeline_name_from_execution_arn,
    load_pipeline_definition_from_s3,
)

FAKE_ROLE_ARN = f"arn:aws:iam::{ACCOUNT_ID}:role/FakeRole"
TEST_REGION_NAME = "us-west-1"


@contextmanager
def setup_s3_pipeline_definition(bucket_name, object_key, pipeline_definition):
    client = boto3.client("s3")
    client.create_bucket(
        Bucket=bucket_name,
        CreateBucketConfiguration={"LocationConstraint": TEST_REGION_NAME},
    )
    client.put_object(
        Body=json.dumps(pipeline_definition),
        Bucket=bucket_name,
        Key=object_key,
    )
    yield

    client.delete_object(Bucket=bucket_name, Key=object_key)
    client.delete_bucket(Bucket=bucket_name)


@pytest.fixture(name="sagemaker_client")
def fixture_sagemaker_client():
    with mock_aws():
        yield boto3.client("sagemaker", region_name=TEST_REGION_NAME)


def create_sagemaker_pipelines(sagemaker_client, pipelines, wait_seconds=0.0):
    responses = []
    for pipeline in pipelines:
        responses += sagemaker_client.create_pipeline(**pipeline)
        sleep(wait_seconds)
    return responses


def test_utils_get_pipeline_from_name_exists():
    fake_pipeline_names = ["APipelineName", "BPipelineName"]
    pipelines = {
        fake_pipeline_name: FakePipeline(
            pipeline_name="BFakePipeline",
            pipeline_display_name="BFakePipeline",
            pipeline_description=" ",
            tags=[],
            parallelism_configuration={},
            pipeline_definition=" ",
            role_arn=FAKE_ROLE_ARN,
            account_id=ACCOUNT_ID,
            region_name=TEST_REGION_NAME,
        )
        for fake_pipeline_name in fake_pipeline_names
    }
    retrieved_pipeline = get_pipeline_from_name(
        pipelines=pipelines, pipeline_name=fake_pipeline_names[0]
    )
    assert retrieved_pipeline == pipelines[fake_pipeline_names[0]]


def test_utils_get_pipeline_from_name_not_exists():
    with pytest.raises(ValidationError):
        _ = get_pipeline_from_name(pipelines={}, pipeline_name="foo")


def test_utils_get_pipeline_name_from_execution_arn():
    expected_pipeline_name = "some-pipeline-name"
    pipeline_execution_arn = (
        f"arn:aws:sagemaker:{TEST_REGION_NAME}:{ACCOUNT_ID}"
        f":pipeline/{expected_pipeline_name}/execution/abc123def456"
    )
    observed_pipeline_name = get_pipeline_name_from_execution_arn(
        pipeline_execution_arn=pipeline_execution_arn
    )
    assert expected_pipeline_name == observed_pipeline_name


def test_utils_get_pipeline_execution_from_arn_not_exists():
    with pytest.raises(ValidationError):
        _ = get_pipeline_execution_from_arn(
            pipelines={},
            pipeline_execution_arn="some/random/non/existent/arn",
        )


def test_utils_arn_formatter():
    expected_arn = (
        f"arn:aws:sagemaker:{TEST_REGION_NAME}:{ACCOUNT_ID}:pipeline/some-pipeline-name"
    )
    observed_arn = arn_formatter(
        _type="pipeline",
        _id="some-pipeline-name",
        region_name=TEST_REGION_NAME,
        account_id=ACCOUNT_ID,
    )
    assert expected_arn == observed_arn


def test_list_pipeline_executions(sagemaker_client):
    fake_pipeline_names = ["APipelineName"]
    pipelines = [
        {
            "PipelineName": fake_pipeline_names[0],
            "RoleArn": FAKE_ROLE_ARN,
            "PipelineDefinition": " ",
        },
    ]
    _ = create_sagemaker_pipelines(sagemaker_client, pipelines)
    _ = sagemaker_client.start_pipeline_execution(PipelineName=fake_pipeline_names[0])
    _ = sagemaker_client.start_pipeline_execution(PipelineName=fake_pipeline_names[0])
    response = sagemaker_client.list_pipeline_executions(
        PipelineName=fake_pipeline_names[0]
    )
    assert len(response["PipelineExecutionSummaries"]) == 2
    assert (
        fake_pipeline_names[0]
        in response["PipelineExecutionSummaries"][0]["PipelineExecutionArn"]
    )


def test_describe_pipeline_definition_for_execution(sagemaker_client):
    fake_pipeline_names = ["APipelineName"]
    pipeline_definition = "some-pipeline-definition"
    pipelines = [
        {
            "PipelineName": fake_pipeline_names[0],
            "RoleArn": FAKE_ROLE_ARN,
            "PipelineDefinition": pipeline_definition,
        },
    ]
    _ = create_sagemaker_pipelines(sagemaker_client, pipelines)
    response = sagemaker_client.start_pipeline_execution(
        PipelineName=fake_pipeline_names[0]
    )
    pipeline_execution_arn = response["PipelineExecutionArn"]
    response = sagemaker_client.describe_pipeline_definition_for_execution(
        PipelineExecutionArn=pipeline_execution_arn
    )
    assert set(response.keys()) == {
        "PipelineDefinition",
        "CreationTime",
        "ResponseMetadata",
    }
    assert response["PipelineDefinition"] == pipeline_definition


def test_list_pipeline_parameters_for_execution(sagemaker_client):
    fake_pipeline_names = ["APipelineName"]
    pipelines = [
        {
            "PipelineName": fake_pipeline_names[0],
            "RoleArn": FAKE_ROLE_ARN,
            "PipelineDefinition": " ",
        },
    ]
    _ = create_sagemaker_pipelines(sagemaker_client, pipelines)
    pipeline_execution_arn = sagemaker_client.start_pipeline_execution(
        PipelineName=fake_pipeline_names[0],
        PipelineParameters=[
            {"Name": "foo", "Value": "bar"},
        ],
    )["PipelineExecutionArn"]

    response = sagemaker_client.list_pipeline_parameters_for_execution(
        PipelineExecutionArn=pipeline_execution_arn
    )
    assert isinstance(response["PipelineParameters"], list)
    assert len(response["PipelineParameters"]) == 1
    assert response["PipelineParameters"][0]["Name"] == "foo"
    assert response["PipelineParameters"][0]["Value"] == "bar"


def test_start_pipeline_execution(sagemaker_client):
    fake_pipeline_names = ["APipelineName"]
    pipelines = [
        {
            "PipelineName": fake_pipeline_names[0],
            "RoleArn": FAKE_ROLE_ARN,
            "PipelineDefinition": " ",
        },
    ]
    _ = create_sagemaker_pipelines(sagemaker_client, pipelines)
    pipeline_execution_arn = sagemaker_client.start_pipeline_execution(
        PipelineName=fake_pipeline_names[0]
    )
    assert fake_pipeline_names[0] in pipeline_execution_arn["PipelineExecutionArn"]


def test_start_pipeline_execution_contains_client_request_token(sagemaker_client):
    if settings.TEST_SERVER_MODE:
        raise SkipTest(
            "Skipping test in server mode due to lack of access to sagemaker_backends."
        )

    fake_pipeline_names = ["APipelineName"]
    pipelines = [
        {
            "PipelineName": fake_pipeline_names[0],
            "RoleArn": FAKE_ROLE_ARN,
            "PipelineDefinition": " ",
        },
    ]
    _ = create_sagemaker_pipelines(sagemaker_client, pipelines)
    pipeline_execution_arn = sagemaker_client.start_pipeline_execution(
        PipelineName=fake_pipeline_names[0]
    )["PipelineExecutionArn"]

    # Verify that client_request_token is stored in FakePipelineExecution object
    assert (
        sagemaker_backends[ACCOUNT_ID][TEST_REGION_NAME]
        .pipelines[fake_pipeline_names[0]]
        .pipeline_executions[pipeline_execution_arn]
        .client_request_token
        != ""
    )


def test_describe_pipeline_execution_not_exists(sagemaker_client):
    pipeline_execution_arn = arn_formatter(
        # random ID (execution ID)
        "pipeline-execution",
        "some-pipeline-name",
        ACCOUNT_ID,
        TEST_REGION_NAME,
    )
    with pytest.raises(botocore.exceptions.ClientError):
        _ = sagemaker_client.describe_pipeline_execution(
            PipelineExecutionArn=pipeline_execution_arn
        )


def test_describe_pipeline_execution(sagemaker_client):
    fake_pipeline_names = ["APipelineName", "BPipelineName"]
    pipelines = [
        {
            "PipelineName": fake_pipeline_name,
            "RoleArn": FAKE_ROLE_ARN,
            "PipelineDefinition": " ",
        }
        for fake_pipeline_name in fake_pipeline_names
    ]
    _ = create_sagemaker_pipelines(sagemaker_client, pipelines)
    response = sagemaker_client.start_pipeline_execution(
        PipelineName=fake_pipeline_names[0]
    )
    _ = sagemaker_client.start_pipeline_execution(PipelineName=fake_pipeline_names[1])
    expected_pipeline_execution_arn = response["PipelineExecutionArn"]
    pipeline_execution_summary = sagemaker_client.describe_pipeline_execution(
        PipelineExecutionArn=response["PipelineExecutionArn"]
    )
    observed_pipeline_execution_arn = pipeline_execution_summary["PipelineExecutionArn"]
    assert observed_pipeline_execution_arn == expected_pipeline_execution_arn


def test_load_pipeline_definition_from_s3():
    if settings.TEST_SERVER_MODE:
        raise SkipTest(
            "Skipping test in server mode due to lack of access to s3_backend."
        )

    bucket_name = "some-bucket-1"
    object_key = "some/object/key.json"
    pipeline_definition = {"key": "value"}

    with mock_aws():
        with setup_s3_pipeline_definition(
            bucket_name,
            object_key,
            pipeline_definition,
        ):
            observed_pipeline_definition = load_pipeline_definition_from_s3(
                pipeline_definition_s3_location={
                    "Bucket": bucket_name,
                    "ObjectKey": object_key,
                },
                account_id=ACCOUNT_ID,
                partition="aws",
            )
    assert observed_pipeline_definition == pipeline_definition


def test_create_pipeline(sagemaker_client):
    fake_pipeline_name = "MyPipelineName"
    response = sagemaker_client.create_pipeline(
        PipelineName=fake_pipeline_name,
        RoleArn=FAKE_ROLE_ARN,
        PipelineDefinition=" ",
    )
    assert isinstance(response, dict)
    assert response["PipelineArn"] == (
        arn_formatter("pipeline", fake_pipeline_name, ACCOUNT_ID, TEST_REGION_NAME)
    )


@pytest.mark.parametrize(
    "create_pipeline_kwargs",
    [
        {"PipelineName": "MyPipelineName", "RoleArn": FAKE_ROLE_ARN},
        {"RoleArn": FAKE_ROLE_ARN, "PipelineDefinition": " "},
        {"PipelineName": "MyPipelineName", "PipelineDefinition": " "},
        {
            "PipelineName": "MyPipelineName",
            "RoleArn": FAKE_ROLE_ARN,
            "PipelineDefinition": " ",
            "PipelineDefinitionS3Location": {"key": "value"},
        },
    ],
)
def test_create_pipeline_invalid_required_kwargs(
    sagemaker_client, create_pipeline_kwargs
):
    with pytest.raises(
        (
            botocore.exceptions.ParamValidationError,
            botocore.exceptions.ClientError,
        )
    ):
        _ = sagemaker_client.create_pipeline(
            **create_pipeline_kwargs,
        )


def test_create_pipeline_duplicate_pipeline_name(sagemaker_client):
    with pytest.raises(botocore.exceptions.ClientError):
        _ = sagemaker_client.create_pipeline(
            PipelineName="APipelineName",
            RoleArn=FAKE_ROLE_ARN,
            PipelineDefinition=" ",
        )
        _ = sagemaker_client.create_pipeline(
            PipelineName="APipelineName",
            RoleArn=FAKE_ROLE_ARN,
            PipelineDefinition=" ",
        )


def test_list_pipelines_none(sagemaker_client):
    response = sagemaker_client.list_pipelines()
    assert isinstance(response, dict)
    assert not response["PipelineSummaries"]


def test_list_pipelines_single(sagemaker_client):
    fake_pipeline_names = ["APipelineName"]
    pipelines = [
        {
            "PipelineName": fake_pipeline_names[0],
            "RoleArn": FAKE_ROLE_ARN,
            "PipelineDefinition": " ",
        },
    ]
    _ = create_sagemaker_pipelines(sagemaker_client, pipelines)

    response = sagemaker_client.list_pipelines()
    assert len(response["PipelineSummaries"]) == 1
    assert response["PipelineSummaries"][0]["PipelineArn"] == (
        arn_formatter("pipeline", fake_pipeline_names[0], ACCOUNT_ID, TEST_REGION_NAME)
    )


def test_list_pipelines_multiple(sagemaker_client):
    fake_pipeline_names = ["APipelineName", "BPipelineName"]
    pipelines = [
        {
            "PipelineName": fake_pipeline_name,
            "RoleArn": FAKE_ROLE_ARN,
            "PipelineDefinition": " ",
        }
        for fake_pipeline_name in fake_pipeline_names
    ]
    _ = create_sagemaker_pipelines(sagemaker_client, pipelines)

    response = sagemaker_client.list_pipelines(
        SortBy="Name",
        SortOrder="Ascending",
    )
    assert len(response["PipelineSummaries"]) == len(fake_pipeline_names)


def test_list_pipelines_sort_name_ascending(sagemaker_client):
    fake_pipeline_names = ["APipelineName", "BPipelineName", "CPipelineName"]
    pipelines = [
        {
            "PipelineName": fake_pipeline_name,
            "RoleArn": FAKE_ROLE_ARN,
            "PipelineDefinition": " ",
        }
        for fake_pipeline_name in fake_pipeline_names
    ]
    _ = create_sagemaker_pipelines(sagemaker_client, pipelines)

    response = sagemaker_client.list_pipelines(
        SortBy="Name",
        SortOrder="Ascending",
    )
    assert response["PipelineSummaries"][0]["PipelineArn"] == (
        arn_formatter("pipeline", fake_pipeline_names[0], ACCOUNT_ID, TEST_REGION_NAME)
    )
    assert response["PipelineSummaries"][-1]["PipelineArn"] == (
        arn_formatter("pipeline", fake_pipeline_names[-1], ACCOUNT_ID, TEST_REGION_NAME)
    )
    assert response["PipelineSummaries"][1]["PipelineArn"] == (
        arn_formatter("pipeline", fake_pipeline_names[1], ACCOUNT_ID, TEST_REGION_NAME)
    )


def test_list_pipelines_sort_creation_time_descending(sagemaker_client):
    fake_pipeline_names = ["APipelineName", "BPipelineName", "CPipelineName"]
    pipelines = [
        {
            "PipelineName": fake_pipeline_name,
            "RoleArn": FAKE_ROLE_ARN,
            "PipelineDefinition": " ",
        }
        for fake_pipeline_name in fake_pipeline_names
    ]
    _ = create_sagemaker_pipelines(sagemaker_client, pipelines, 1.0)

    response = sagemaker_client.list_pipelines(
        SortBy="CreationTime",
        SortOrder="Descending",
    )
    assert response["PipelineSummaries"][0]["PipelineArn"] == (
        arn_formatter("pipeline", fake_pipeline_names[-1], ACCOUNT_ID, TEST_REGION_NAME)
    )
    assert response["PipelineSummaries"][1]["PipelineArn"] == (
        arn_formatter("pipeline", fake_pipeline_names[1], ACCOUNT_ID, TEST_REGION_NAME)
    )
    assert response["PipelineSummaries"][2]["PipelineArn"] == (
        arn_formatter("pipeline", fake_pipeline_names[0], ACCOUNT_ID, TEST_REGION_NAME)
    )


def test_list_pipelines_max_results(sagemaker_client):
    fake_pipeline_names = ["APipelineName", "BPipelineName", "CPipelineName"]
    pipelines = [
        {
            "PipelineName": fake_pipeline_name,
            "RoleArn": FAKE_ROLE_ARN,
            "PipelineDefinition": " ",
        }
        for fake_pipeline_name in fake_pipeline_names
    ]
    _ = create_sagemaker_pipelines(sagemaker_client, pipelines)

    response = sagemaker_client.list_pipelines(MaxResults=2)
    assert len(response["PipelineSummaries"]) == 2


def test_list_pipelines_next_token(sagemaker_client):
    fake_pipeline_names = ["APipelineName"]
    pipelines = [
        {
            "PipelineName": fake_pipeline_names[0],
            "RoleArn": FAKE_ROLE_ARN,
            "PipelineDefinition": " ",
        },
    ]
    _ = create_sagemaker_pipelines(sagemaker_client, pipelines)

    response = sagemaker_client.list_pipelines(NextToken="0")
    assert len(response["PipelineSummaries"]) == 1


def test_list_pipelines_pipeline_name_prefix(sagemaker_client):
    fake_pipeline_names = ["APipelineName", "BPipelineName", "CPipelineName"]
    pipelines = [
        {
            "PipelineName": fake_pipeline_name,
            "RoleArn": FAKE_ROLE_ARN,
            "PipelineDefinition": " ",
        }
        for fake_pipeline_name in fake_pipeline_names
    ]
    _ = create_sagemaker_pipelines(sagemaker_client, pipelines)

    response = sagemaker_client.list_pipelines(PipelineNamePrefix="APipe")
    assert len(response["PipelineSummaries"]) == 1
    assert response["PipelineSummaries"][0]["PipelineName"] == "APipelineName"

    response = sagemaker_client.list_pipelines(PipelineNamePrefix="Pipeline")
    assert len(response["PipelineSummaries"]) == 3


def test_list_pipelines_created_after(sagemaker_client):
    fake_pipeline_names = ["APipelineName", "BPipelineName", "CPipelineName"]
    pipelines = [
        {
            "PipelineName": fake_pipeline_name,
            "RoleArn": FAKE_ROLE_ARN,
            "PipelineDefinition": " ",
        }
        for fake_pipeline_name in fake_pipeline_names
    ]
    _ = create_sagemaker_pipelines(sagemaker_client, pipelines)

    created_after_str = "2099-12-31 23:59:59"
    try:
        response = sagemaker_client.list_pipelines(CreatedAfter=created_after_str)
    except OverflowError:
        pytest.skip("This test requires 64-bit time_t")
    assert not response["PipelineSummaries"]

    created_after_datetime = datetime.strptime(created_after_str, "%Y-%m-%d %H:%M:%S")
    response = sagemaker_client.list_pipelines(CreatedAfter=created_after_datetime)
    assert not response["PipelineSummaries"]

    created_after_timestamp = datetime.timestamp(created_after_datetime)
    response = sagemaker_client.list_pipelines(CreatedAfter=created_after_timestamp)
    assert not response["PipelineSummaries"]


def test_list_pipelines_created_before(sagemaker_client):
    fake_pipeline_names = ["APipelineName", "BPipelineName", "CPipelineName"]
    pipelines = [
        {
            "PipelineName": fake_pipeline_name,
            "RoleArn": FAKE_ROLE_ARN,
            "PipelineDefinition": " ",
        }
        for fake_pipeline_name in fake_pipeline_names
    ]
    _ = create_sagemaker_pipelines(sagemaker_client, pipelines)

    created_before_str = "2000-12-31 23:59:59"
    response = sagemaker_client.list_pipelines(CreatedBefore=created_before_str)
    assert not response["PipelineSummaries"]

    created_before_datetime = datetime.strptime(created_before_str, "%Y-%m-%d %H:%M:%S")
    response = sagemaker_client.list_pipelines(CreatedBefore=created_before_datetime)
    assert not response["PipelineSummaries"]

    created_before_timestamp = datetime.timestamp(created_before_datetime)
    response = sagemaker_client.list_pipelines(CreatedBefore=created_before_timestamp)
    assert not response["PipelineSummaries"]


@pytest.mark.parametrize(
    "list_pipelines_kwargs",
    [
        {"MaxResults": 200},
        {"NextToken": "some-invalid-next-token"},
        {"SortOrder": "some-invalid-sort-order"},
        {"SortBy": "some-invalid-sort-by"},
    ],
)
def test_list_pipelines_invalid_values(sagemaker_client, list_pipelines_kwargs):
    with pytest.raises(botocore.exceptions.ClientError):
        _ = sagemaker_client.list_pipelines(**list_pipelines_kwargs)


def test_delete_pipeline_exists(sagemaker_client):
    fake_pipeline_names = ["APipelineName", "BPipelineName", "CPipelineName"]
    pipelines = [
        {
            "PipelineName": fake_pipeline_name,
            "RoleArn": FAKE_ROLE_ARN,
            "PipelineDefinition": " ",
        }
        for fake_pipeline_name in fake_pipeline_names
    ]
    _ = create_sagemaker_pipelines(sagemaker_client, pipelines)
    pipeline_name_delete, pipeline_names_remain = (
        fake_pipeline_names[0],
        fake_pipeline_names[1:],
    )

    response = sagemaker_client.delete_pipeline(PipelineName=pipeline_name_delete)
    assert response["PipelineArn"].endswith(pipeline_name_delete)

    response = sagemaker_client.list_pipelines(PipelineNamePrefix=pipeline_name_delete)
    assert not response["PipelineSummaries"]

    response = sagemaker_client.list_pipelines()
    pipeline_names_exist = [
        pipeline["PipelineName"] for pipeline in response["PipelineSummaries"]
    ]
    assert set(pipeline_names_remain) == set(pipeline_names_exist)


def test_delete_pipeline_not_exists(sagemaker_client):
    with pytest.raises(botocore.exceptions.ClientError):
        _ = sagemaker_client.delete_pipeline(PipelineName="some-pipeline-name")


def test_update_pipeline_not_exists(sagemaker_client):
    with pytest.raises(botocore.exceptions.ClientError):
        _ = sagemaker_client.update_pipeline(PipelineName="some-pipeline-name")


def test_update_pipeline_invalid_kwargs(sagemaker_client):
    pipeline_name = "APipelineName"
    pipeline = {
        "PipelineName": pipeline_name,
        "RoleArn": FAKE_ROLE_ARN,
        "PipelineDefinition": " ",
    }
    _ = create_sagemaker_pipelines(sagemaker_client, [pipeline])

    with pytest.raises(botocore.exceptions.ParamValidationError):
        sagemaker_client.update_pipeline(
            PipelineName=pipeline_name,
            **{"InvalidKwarg": "some-value"},
        )


def test_update_pipeline_no_update(sagemaker_client):
    pipeline_name = "APipelineName"
    pipeline = {
        "PipelineName": pipeline_name,
        "RoleArn": FAKE_ROLE_ARN,
        "PipelineDefinition": " ",
    }
    _ = create_sagemaker_pipelines(sagemaker_client, [pipeline])

    response = sagemaker_client.update_pipeline(PipelineName=pipeline_name)
    assert response["PipelineArn"] == (
        arn_formatter("pipeline", pipeline_name, ACCOUNT_ID, TEST_REGION_NAME)
    )
    response = sagemaker_client.list_pipelines()
    assert response["PipelineSummaries"][0]["PipelineName"] == pipeline_name


def test_update_pipeline_add_attribute(sagemaker_client):
    pipeline_name = "APipelineName"
    pipeline_display_name_update = "APipelineDisplayName"

    pipeline = {
        "PipelineName": pipeline_name,
        "RoleArn": FAKE_ROLE_ARN,
        "PipelineDefinition": " ",
    }
    _ = create_sagemaker_pipelines(sagemaker_client, [pipeline])
    response = sagemaker_client.list_pipelines()
    assert response["PipelineSummaries"][0]["PipelineDisplayName"] == pipeline_name

    _ = sagemaker_client.update_pipeline(
        PipelineName=pipeline_name,
        PipelineDisplayName=pipeline_display_name_update,
    )
    response = sagemaker_client.list_pipelines()
    assert response["PipelineSummaries"][0]["PipelineDisplayName"] == (
        pipeline_display_name_update
    )
    assert len(response["PipelineSummaries"][0]) == 6


def test_update_pipeline_update_change_attribute(sagemaker_client):
    pipeline_name = "APipelineName"
    role_arn_update = f"{FAKE_ROLE_ARN}Test"
    pipeline = {
        "PipelineName": pipeline_name,
        "RoleArn": FAKE_ROLE_ARN,
        "PipelineDefinition": " ",
    }
    _ = create_sagemaker_pipelines(sagemaker_client, [pipeline])

    _ = sagemaker_client.update_pipeline(
        PipelineName=pipeline_name,
        RoleArn=role_arn_update,
    )
    response = sagemaker_client.list_pipelines()
    assert response["PipelineSummaries"][0]["RoleArn"] == role_arn_update
    assert len(response["PipelineSummaries"][0]) == 6


def test_describe_pipeline_not_exists(sagemaker_client):
    with pytest.raises(botocore.exceptions.ClientError):
        _ = sagemaker_client.describe_pipeline(PipelineName="some-pipeline-name")


@pytest.mark.parametrize(
    "pipeline,expected_response_length",
    [
        (
            {
                "PipelineName": "APipelineName",
                "RoleArn": FAKE_ROLE_ARN,
                "PipelineDefinition": " ",
            },
            11,
        ),
        (
            {
                "PipelineName": "BPipelineName",
                "RoleArn": FAKE_ROLE_ARN,
                "PipelineDefinition": " ",
                "PipelineDescription": "some pipeline description",
            },
            12,
        ),
    ],
)
def test_describe_pipeline_exists(sagemaker_client, pipeline, expected_response_length):
    _ = create_sagemaker_pipelines(sagemaker_client, [pipeline])
    response = sagemaker_client.describe_pipeline(PipelineName=pipeline["PipelineName"])
    assert len(response) == expected_response_length
